import numpy as np
import gc
import os

import numpy as np

from gennet.utils.hase.hdgwas.hdregression import HASE, A_covariates, A_tests, B_covariates, C_matrix, A_inverse, B4
from gennet.utils.hase.hdgwas.tools import study_indexes, Timer, merge_genotype


def haseregression(phen, gen, cov, mapper, Analyser, maf, intercept=True, interaction=None):
    g = tuple([i.folder._data for i in gen])

    row_index, ids = study_indexes(phenotype=phen.folder._data,
                                   genotype=g,
                                   covariates=cov.folder._data)

    if mapper is not None:
        SNP = [0, 0, mapper.n_keys]
    else:
        SNP = [0, 0, 'unknown']

    covariates = cov.get_next(index=row_index[2])
    a_cov = A_covariates(covariates, intercept=intercept)

    while True:
        gc.collect()
        if mapper is not None:
            if mapper.cluster == 'n':
                SNPs_index, keys = mapper.get()
            else:
                ch = mapper.chunk_pop()
                if ch is None:
                    SNPs_index = None
                    break
                SNPs_index, keys = mapper.get(chunk_number=ch)
            if isinstance(SNPs_index, type(None)):
                break
            Analyser.rsid = keys
        else:
            SNPs_index = None

        with Timer() as t:
            genotype = merge_genotype(gen, SNPs_index, mapper)
        print(('time to read and merge genotype {}s'.format(t.secs)))
        gc.collect()
        if genotype is None:
            print('All genotype processed!')
            break
        SNP[0] += genotype.shape[0]
        genotype = genotype[:, row_index[0]]

        if mapper is None:
            Analyser.rsid = np.array(list(range(genotype.shape[0])))

        MAF = np.mean(genotype, axis=1) / 2
        STD = np.std(genotype, axis=1)

        if maf != 0:

            filter = (MAF > maf) & (MAF < 1 - maf) & (MAF != 0.5)
            genotype = genotype[filter, :]
            Analyser.MAF = MAF[filter]
            Analyser.rsid = Analyser.rsid[filter]

            if genotype.shape[0] == 0:
                print('NO SNPs > MAF')
                continue

        else:
            Analyser.MAF = MAF

        SNP[1] += genotype.shape[0]

        while True:
            phenotype = phen.get_next(index=row_index[1])

            if isinstance(phenotype, type(None)):
                phen.folder.processed = 0
                print('All phenotypes processed!')
                break

            if phen.permutation:
                np.random.shuffle(phenotype)

            b_cov = B_covariates(covariates, phenotype, intercept=intercept)

            C = C_matrix(phenotype)

            if interaction is not None:
                pass

            a_test = A_tests(covariates, genotype, intercept=intercept)
            a_inv = A_inverse(a_cov, a_test)

            N_con = a_inv.shape[1] - 1

            DF = (phenotype.shape[0] - a_inv.shape[1])

            b4 = B4(phenotype, genotype)

            t_stat, SE = HASE(b4, a_inv, b_cov, C, N_con, DF)
            print(('Read {}, processed {}, total {}'.format(SNP[0], SNP[1], SNP[2])))
            Analyser.t_stat = t_stat
            Analyser.SE = SE
            if mapper is not None and mapper.cluster == 'y':
                Analyser.cluster = True
                Analyser.chunk = ch
                Analyser.node = mapper.node[1]
            if phen.permutation:
                Analyser.permutation = True
            Analyser.save_result(phen.folder._data.names[phen.folder._data.start:phen.folder._data.finish])
            t_stat = None
            Analyser.t_stat = None
            del b4
            del C
            del b_cov
            del a_inv
            del a_test
            del t_stat
            gc.collect()

    if Analyser.cluster:
        np.save(os.path.join(Analyser.out, str(Analyser.node) + '_node_RSID.npy'), Analyser.rsid_dic)



import sys
import os
import numpy as np
import time
import tqdm
import pandas as pd
import tensorflow as tf
import multiprocessing

sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

import shap
import interpretation.DFIM as DFIM
        
from interpretation.weight_importance import make_importance_values_input
from interpretation.NID import Get_weight_tsang, GenNet_pairwise_interactions_topn

from gennet.utils.Train_network import load_trained_network
from gennet.utils.Create_network import remove_batchnorm_model, remove_cov
from gennet.utils.Dataloader import EvalGenerator

def interpret(args):
    if args.type == 'get_weight_scores':
        get_weight_scores(args)
    elif args.type == 'NID':
        get_NID_scores(args)
    elif args.type == 'DeepExplain':
        get_DeepExplainer_scores(args)
    elif args.type == 'RLIPP':
        get_RLIPP_scores(args)
    elif args.type == 'DFIM':
        get_DFIM_scores(args)
    elif args.type == 'PathExplain':
        get_pathexplain_scores(args)
    else:
        print("invalid type:", args.type)
        exit()


def get_weight_scores(args):
    model, masks = load_trained_network(args)

    if os.path.exists(args.resultpath + "/weight_importance.npy"):
        print('weight Done')
    else:
        weight_importance = make_importance_values_input(model, masks=masks)
        np.save(args.resultpath + "/weight_importance.npy", weight_importance)


def get_DeepExplainer_scores(args):
    tf.compat.v1.disable_eager_execution()

    print("Interpreting with DFIM:")

    model, masks = load_trained_network(args)

    xval, yval= EvalGenerator(datapath=args.path, genotype_path=args.genotype_path, batch_size=64,
                                          setsize=-1, one_hot=args.onehot,
                                          inputsize=-1, evalset="validation").get_data(sample_pat=args.num_sample_pat)
    xtest, ytest = EvalGenerator(datapath=args.path, genotype_path=args.genotype_path, batch_size=64,
                                          setsize=-1, one_hot=args.onehot,
                                          inputsize=-1, evalset="test").get_data(sample_pat= args.num_sample_pat)


    print("Loaded the data")
    
    model = remove_batchnorm_model(model, masks, keep_cov=False)

    print("compile")
    model.compile(optimizer=tf.keras.optimizers.Adam(learning_rate=1e-3),
              loss=tf.keras.losses.BinaryCrossentropy())

    xval = xval[0]
    xtest = xtest[0]

    explainer  = shap.DeepExplainer((model.input, model.output), xval)
    print("Created explainer")

    if os.path.exists( args.resultpath+ "/DeepExplain_test.npy"):
        shap_values = np.load(args.resultpath + "/DeepExplain_test.npy")
    else:
        max_axis = (0,2) if args.onehot else 0
        shap_values = np.max(explainer.shap_values(xtest)[0], axis=max_axis)
        np.save(args.resultpath + "/DeepExplain_test.npy", shap_values)
    
    print("DeepExplain test results saved to " + args.resultpath)  


def get_NID_scores(args):
    
    print("Interpreting with NID:")
    model, masks = load_trained_network(args)

    mask = masks[0]

    if args.layer == None:
        if args.onehot == 1:
            interp_layer = 3
        else:
            interp_layer = 2
    else:
        interp_layer = args.layer

    print("Interrpeting layer", interp_layer)
    if os.path.exists(args.resultpath + "/NID.csv"):
        print('RLIPP Done')
        interaction_ranking = pd.read_csv(args.resultpath + "/NID.csv")
    else:
        print("Obtaining the weights")
        w_in, w_out = Get_weight_tsang(model, interp_layer, masks)
        print("Computing interactions")

        pairwise_interactions_dfs = []
        for filter in range(w_in.shape[1]):  # for all the filters
            pairwise_interactions = GenNet_pairwise_interactions_topn(w_in[:,filter] ,w_out[:,filter], mask, n="auto")
            pairwise_interactions_dfs.append(pairwise_interactions)
        
        interaction_ranking = pd.concat(pairwise_interactions_dfs)
        interaction_ranking = interaction_ranking.sort_values("strength", ascending =False)
        interaction_ranking.to_csv(args.resultpath + "/NID.csv")
        print("NID results are saved in", args.resultpath + "/NID.csv")
  
    return interaction_ranking



def get_RLIPP_scores(args):
    print("not implemented yet")



def get_DFIM_scores(args):
    tf.compat.v1.disable_eager_execution()

    print("Interpreting with DFIM:")
    
    num_snps_to_eval = args.num_eval if hasattr(args, 'num_eval') else 100

    model, masks = load_trained_network(args)



    xval, yval= EvalGenerator(datapath=args.datapath, genotype_path=args.genotype_path, batch_size=64,
                                          setsize=-1, one_hot=args.onehot,
                                          inputsize=-1, evalset="validation").get_data(sample_pat=args.num_sample_pat)
    xtest, ytest = EvalGenerator(datapath=args.datapath, genotype_path=args.genotype_path, batch_size=64,
                                          setsize=-1, one_hot=args.onehot,
                                          inputsize=-1, evalset="test").get_data(sample_pat=args.num_sample_pat)


    print("Loaded the data")
    
    model = remove_batchnorm_model(model, masks, keep_cov=False)

    print("compile")
    model.compile(optimizer=tf.keras.optimizers.Adam(learning_rate=1e-3),
              loss=tf.keras.losses.BinaryCrossentropy())

    xval = xval[0]
    xtest = xtest[0]

    explainer  = shap.DeepExplainer((model.input, model.output), xval)
    print("Created explainer")

    if os.path.exists( args.resultpath+ "/shap_test.npy"):
        shap_values = np.load(args.resultpath + "/shap_test.npy")
    else:
        max_axis = (0,2) if args.onehot else 0
        shap_values = np.max(explainer.shap_values(xtest)[0], axis=max_axis)
        np.save(args.resultpath + "/shap_test.npy", shap_values)
    
    print("Find most important SNPs..")
        
    
    snp_num_eval = min(num_snps_to_eval, shap_values.shape[0])
    snp_index = np.argsort(shap_values)[::-1][:snp_num_eval]

    part_suffix = ""
    if (args.start_rank == 0) and (args.end_rank == 0):
        part_suffix = ""
    else:
        part_suffix = "_" + str(args.start_rank) + "_" + str(args.end_rank) + "_"
        snp_index = snp_index[args.start_rank:args.end_rank] 

    print("Evaluate the following variants", snp_index)

    print("Start DFIM for the", snp_num_eval, "most important SNPs -> see ", args.resultpath + "/DFIM_loc_not_perturbed_"
          +str(part_suffix)+".npy", "when finished" )

    perturbed_values, max_not_perturbed, loc_not_perturbed= DFIM.DFIM_test_index(explainer, xtest, snp_index)
    np.save(args.resultpath + "/DFIM_not_perturbed_"+str(part_suffix)+".npy", max_not_perturbed)
    np.save(args.resultpath + "/DFIM_loc_not_perturbed_"+str(part_suffix)+".npy", loc_not_perturbed)
    np.save(args.resultpath + "/DFIM_perturbed_"+str(part_suffix)+".npy", perturbed_values)
    time_DFIM = time.time()
    print("results saved to", args.resultpath)
    
    return
    
def get_pathexplain_scores(args):
    from path_explain import PathExplainerTF

    num_snps_to_eval = args.num_eval if hasattr(args, 'num_eval') else 100

    model, masks = load_trained_network(args)

    print("evaluate over", args.num_sample_pat, "exomples")

    xval, yval= EvalGenerator(datapath=args.path, genotype_path=args.genotype_path, batch_size=64,
                                          setsize=-1, one_hot=args.onehot,
                                          inputsize=-1, evalset="validation").get_data(sample_pat=args.num_sample_pat)
    xtest, ytest = EvalGenerator(datapath=args.path, genotype_path=args.genotype_path, batch_size=64,
                                          setsize=-1, one_hot=args.onehot,
                                          inputsize=-1, evalset="test").get_data(sample_pat= args.num_sample_pat)

    model = remove_cov(model, masks)
    
    xval = xval[0]
    xtest = xtest[0]
    print("Shapes",xval.shape, xtest.shape)

    explainer = PathExplainerTF(model)
    n_top_values = min(num_snps_to_eval, xtest.shape[1])

    if os.path.exists( args.resultpath+ "/pathexplain_snp_index.npy"):
        SNP_indices = np.load(args.resultpath + "/pathexplain_snp_index.npy")
    else:
        attributions = explainer.attributions(xtest.astype(np.float32), xval.astype(np.float32),
                                batch_size=100, num_samples=args.num_sample_pat,
                                use_expectation=True, output_indices=[0] * len(xtest),
                                verbose=True)
        if args.onehot:
            attr_matrix = np.max(abs(attributions), axis=(0,2))
        else:
            attr_matrix = np.max(abs(attributions), axis=0)

        SNP_indices = np.argsort(attr_matrix)[::-1]
        np.save(args.resultpath + "/pathexplain_snp_index", SNP_indices)


    SNP_indices = SNP_indices[:n_top_values]
    
    part_suffix = ""
    if (args.start_rank == 0) and (args.end_rank == 0):
        part_suffix = ""
    else:
        part_suffix = "_" + str(args.start_rank) + "_" + str(args.end_rank) + "_"
        SNP_indices = SNP_indices[args.start_rank:args.end_rank] 

    print("Evaluate the following variants", SNP_indices)

    interactions = []

    if args.onehot:
        print("using one hot")
        for SNP_index in tqdm.tqdm(SNP_indices):
            interaction_onehot = []
            for one_hot_encoding in range(3):
                interaction_ = explainer.interactions(xtest.astype(np.float32), xval.astype(np.float32),
                        batch_size=100, num_samples=100,
                        use_expectation=True, output_indices=[0] * len(xtest),
                        verbose=True, interaction_index=[SNP_index, one_hot_encoding])
                interaction__ = np.max(abs(interaction_), axis=(0,2))
                interaction_onehot.append(interaction__)

            interactions.append(np.max(np.stack(interaction_onehot), axis=0))

        interactions = np.stack(interactions)
    else:
        for SNP_index in tqdm.tqdm(SNP_indices):               

            interaction = explainer.interactions(xtest.astype(np.float32), xval.astype(np.float32),
                        batch_size=100, num_samples=args.num_sample_pat,
                        use_expectation=True, output_indices=[0] * len(xtest),
                        verbose=True, interaction_index=[SNP_index])
            interaction = np.max(abs(interaction), axis=(0))
            interactions.append(interaction)
        interactions = np.stack(interactions)

    time_pathexplain = time.time()
    print("Saving results...")

    np.save(args.resultpath + "/pathexplain_inter" + part_suffix, interactions)



import tensorflow as tf
import tensorflow.keras.backend as K
import scipy.sparse as sp


class PerVariantNormalization(tf.keras.layers.Layer):
    def __init__(self, momentum=0.99, epsilon=1e-6, **kwargs):
        super(PerVariantNormalization, self).__init__(**kwargs)
        self.momentum = momentum
        self.epsilon = epsilon

    def build(self, input_shape):
        # Initialize mean and variance for each feature (genetic variant)
        self.mean = self.add_weight(name='mean',
                                    shape=(input_shape[-1],),
                                    initializer='zeros',
                                    trainable=False)
        self.variance = self.add_weight(name='variance',
                                        shape=(input_shape[-1],),
                                        initializer='ones',
                                        trainable=False)

    def call(self, inputs, training=None):
        if training:
            # Compute mean and variance for the batch
            batch_mean, batch_variance = tf.nn.moments(inputs, axes=[0], keepdims=False)

            # Update the running mean and variance
            new_mean = self.momentum * self.mean + (1 - self.momentum) * batch_mean
            new_variance = self.momentum * self.variance + (1 - self.momentum) * batch_variance

            self.mean.assign(new_mean)
            self.variance.assign(new_variance)

        # Normalize each feature (genetic variant) using the running mean and variance
        x = (inputs - self.mean) / tf.sqrt(self.variance + self.epsilon)
        return x

    def update_mean_and_variance(self, new_mean, new_variance):
        # Update the mean and variance for each genetic variant
        self.mean.assign(new_mean)
        self.variance.assign(new_variance)


class ConnectedNormalization(tf.keras.layers.Layer):
    def __init__(self, connectivity_matrix, momentum=0.99, epsilon=1e-6, **kwargs):
        super(ConnectedNormalization, self).__init__(**kwargs)
        self.momentum = momentum
        self.epsilon = epsilon
        # Convert the provided COO matrix to TensorFlow SparseTensor
        self.connectivity_matrix = tf.sparse.SparseTensor(
            indices=np.array([connectivity_matrix.row, connectivity_matrix.col]).T,
            values=connectivity_matrix.data,
            dense_shape=connectivity_matrix.shape
        )

    def build(self, input_shape):
        # Number of neurons (columns of the connectivity matrix)
        num_neurons = self.connectivity_matrix.dense_shape[1]

        # Initialize mean and variance for each neuron
        self.mean = self.add_weight(name='mean',
                                    shape=(num_neurons,),
                                    initializer='zeros',
                                    trainable=False)
        self.variance = self.add_weight(name='variance',
                                        shape=(num_neurons,),
                                        initializer='ones',
                                        trainable=False)

    def call(self, inputs, training=None):
        # Normalize based on connectivity
        # Each neuron's statistics are based on connected inputs

        # Step 1: Calculate weighted sums of inputs for each neuron
        weighted_sums = tf.sparse.sparse_dense_matmul(self.connectivity_matrix, inputs, adjoint_a=True)

        # Step 2: Count connections for each neuron to compute means
        connection_counts = tf.reduce_sum(tf.sparse.to_dense(self.connectivity_matrix), axis=0)
        means = weighted_sums / connection_counts

        # Step 3: Compute variances
        squared_inputs = tf.pow(inputs, 2)
        weighted_squared_sums = tf.sparse.sparse_dense_matmul(self.connectivity_matrix, squared_inputs, adjoint_a=True)
        variances = weighted_squared_sums / connection_counts - tf.pow(means, 2)

        if training:
            # Update the running mean and variance
            self.mean.assign(self.momentum * self.mean + (1 - self.momentum) * means)
            self.variance.assign(self.momentum * self.variance + (1 - self.momentum) * variances)

        # Step 4: Normalize inputs
        normalized_inputs = (inputs - tf.sparse.sparse_dense_matmul(self.connectivity_matrix, tf.expand_dims(self.mean, 1))) / \
                            tf.sqrt(tf.sparse.sparse_dense_matmul(self.connectivity_matrix, tf.expand_dims(self.variance, 1)) + self.epsilon)

        return normalized_inputs



import os
import sys
import pandas as pd
import shutil
from os.path import dirname, abspath
import os
sys.path.insert(1, os.getcwd())
from gennet.utils.Interpret import interpret

# TODO: add test without covariates
# TODO add test with covariates for regression + classification
# TODO add test with multiple genotype files.
# test randomnesss after .. epoch shuffles.
# ToDO add test for each file.

class ArgparseSimulator():
    def __init__(self,
                resultpath = "//trinity/home/avanhilten/repositories/Dev/GenNet/results/GenNet_experiment_16_/",
                genotype_path = "examples/example_classification/",
                type = "NID",
                layer = None,
                ):

        self.resultpath = resultpath
        self.type = type
        self.layer = layer
        self.genotype_path = genotype_path

if __name__ == "__main__":     
    args = ArgparseSimulator(type="DFIM")

    print(args.type)
    print("done")
    interpret(args)



import glob
import inspect
import time
from collections import OrderedDict
import numpy as np
import h5py
import pandas as pd
from scipy import stats

from gennet.utils.hase.hdgwas.hash import *


def timer(func):
    def f(*args, **kwargs):
        with Timer(verbose=True) as t:
            return func(*args, **kwargs)

    return f


class Timer(object):
    def __init__(self, verbose=False):
        self.verbose = verbose

    def __enter__(self):
        self.start = time.time()
        return self

    def __exit__(self, *args):
        self.end = time.time()
        self.secs = self.end - self.start
        self.msecs = self.secs * 1000  # millisecs
        if self.verbose:
            print('elapsed time: %f ms' % self.msecs)


def timing(f):
    def wrap(*args, **kwargs):
        time1 = time.time()
        ret = f(*args, **kwargs)
        time2 = time.time()
        print('%s function took %0.3f s' % (f.__name__, (time2 - time1)))
        return ret

    return wrap


def save_parameters(f):
    def wrap(*args, **kwargs):
        args_name = inspect.getargspec(f)[0]
        args_dict = dict(zip(args_name, args))
        np.save('{}.npy'.format(f.__name__), args_dict)
        ret = f(*args, **kwargs)
        return ret

    return wrap


def check_np():
    for i in ['atlas_blas_info', 'openblas_info', 'lapack_opt_info',
              'atlas_info', 'lapack_mkl_info', 'blas_mkl_info',
              'atlas_blas_info', 'mkl_info']:
        try:
            if len(getattr(np.__config__, i)) > 0:
                print(('Numpy linked to {}'.format(i)))
                return 0
        except:
            continue

    raise ValueError('NO BLAS/LAPACK/MKL found!'
                     'To restart with slow speed use -np False')


########################################################################
########################################################################
########################################################################

class Analyser(object):

    def __init__(self, name):
        self.name = name


class HaseAnalyser(Analyser):

    def __init__(self):
        self.name = None
        self.t_stat = None
        self.p_value = None
        self.betas = None
        self.SE = None
        self.DF = None
        self.n_studies = None
        self.probes_path = {}
        self.results = OrderedDict()
        self.split_size = None
        self.threshold = None
        self.result_path = None
        self.result_index = 1
        self.out = None
        self.MAF = None
        self.rsid = None
        self.reference = None
        self.cluster = False
        self.permutation = False
        self.rsid_dic = {}
        self.result_folder = None
        self.result_dump_size = 3
        self.file_number = None

    def summary(self):

        if isinstance(self.result_path, type(None)):
            raise ValueError('Please set result pathway to Analyser!')

        if isinstance(self.DF, type(None)):
            print('DF is not defined. Forced to use z_score statistics!')

        if self.result_folder is None:
            self.result_folder = [i for i in glob.glob(os.path.join(self.result_path, '*.npy')) if 'RSID' not in i]

        self.results['RSID'] = np.array([])
        self.results['p_value'] = np.array([])
        self.results['t-stat'] = np.array([])
        self.results['phenotype'] = np.array([])
        self.results['SE'] = np.array([])
        self.results['MAF'] = np.array([])
        self.results['BETA'] = np.array([])

        files = []

        if self.file_number is not None:
            try:
                d = np.load(os.path.join(self.result_path, '{}result.npy'.format(self.file_number))).item()
            except:
                print("Can't read {}".format(i))

            p_value = self.get_p_value(d['t-stat'], df=self.DF)
            self.results['t-stat'] = np.append(self.results['t-stat'], d['t-stat'].flatten())
            self.results['SE'] = np.append(self.results['SE'], d['SE'].flatten())
            self.results['RSID'] = np.append(self.results['RSID'], d['index'])
            self.results['phenotype'] = np.append(self.results['phenotype'], d['phenotype'])
            self.results['MAF'] = np.append(self.results['MAF'], d['MAF'])
            self.results['p_value'] = np.append(self.results['p_value'], p_value.flatten())
            self.results['BETA'] = np.append(self.results['BETA'], d['t-stat'].flatten() * d['SE'].flatten())

        else:

            for i in range(self.result_dump_size):
                try:
                    files.append(self.result_folder.pop())
                except:
                    break
            if len(files) != 0:
                for i in files:
                    print(i)
                    try:
                        d = np.load(i).item()
                    except:
                        print("Can't read {}".format(i))
                        continue
                    p_value = self.get_p_value(d['t-stat'], df=self.DF)
                    self.results['t-stat'] = np.append(self.results['t-stat'], d['t-stat'].flatten())
                    self.results['SE'] = np.append(self.results['SE'], d['SE'].flatten())
                    self.results['RSID'] = np.append(self.results['RSID'], d['index'])
                    self.results['phenotype'] = np.append(self.results['phenotype'], d['phenotype'])
                    self.results['MAF'] = np.append(self.results['MAF'], d['MAF'])
                    self.results['p_value'] = np.append(self.results['p_value'], p_value.flatten())
                    self.results['BETA'] = np.append(self.results['BETA'], d['t-stat'].flatten() * d['SE'].flatten())
            else:
                self.results = None

    def get_p_value(self, t_stat, df=None):
        if df is None:
            return stats.norm.sf(np.abs(t_stat)) * 2
        else:
            return stats.t.sf(np.abs(t_stat), df) * 2

    def save_result(self, phen_names):

        t_threshold = self.threshold
        save_path = self.out

        mask = ([], [], [])

        if t_threshold:
            mask = np.where(np.abs(self.t_stat) > t_threshold)
        else:
            mask = np.where(np.abs(self.t_stat) > 0)

        if (len(mask[0]) != 0):
            print(('Saving results to {}'.format(save_path)))
            t_save = self.t_stat[mask[0], mask[1], mask[2]]
            se = self.SE[mask[0], mask[1], mask[2]]
            result = {'phenotype': phen_names[mask[2]], 't-stat': t_save, 'index': self.rsid[mask[0]], 'SE': se,
                      'MAF': self.MAF[mask[0]]}

            if not self.cluster:
                np.save(os.path.join(save_path, str(self.result_index) + 'result.npy'), result)
            else:
                if self.permutation:
                    result = self.t_stat
                self.rsid_dic[str(self.chunk[0]) + '_' + str(self.chunk[1])] = self.rsid
                np.save(os.path.join(save_path, 'node' + str(self.node) + '_' + str(self.chunk[0]) + '_' + str(
                    self.chunk[1]) + '_' + str(self.result_index) + 'result.npy'), result)
        self.result_index += 1


########################################################################
########################################################################
########################################################################
class Log(object):  # TODO (low) write this class

    def __init__(self):
        pass


########################################################################
########################################################################
########################################################################
class Checker(object):  # TODO (mid) finish or remove this class

    def __init__(self):
        self.mode = None
        self.name = None
        self.args = None

    def check(self, args, mode=None):

        if mode == 'converting':
            return self.converting(args)
        elif mode == 'encoding':
            self.encoding(args)
        elif mode == 'single-meta':
            self.single_meta(args)
        elif mode == 'meta-stage':
            self.meta_stage(args)
        elif mode == 'regression':
            self.regression(args)

    def system_check(self, args):
        pass

    def converting(self, args):

        if len(args.study_name) > 1:
            raise ValueError('Should be only one study name!')

        if len(args.genotype) > 1:
            raise ValueError('Should be only one directory for genotype!')

        if isinstance(args.study_name, type(None)):
            raise ValueError('study name is not defined')

        args.genotype = args.genotype[0]
        if not os.path.isdir(args.genotype):
            raise ValueError('{} is not a directory!'.format(args.genotype))

    def encoding(self, args):

        if len(args.study_name) > 1:
            raise ValueError('Should be only one study name!')

        if len(args.genotype) > 1:
            raise ValueError('Should be only one directory for genotype!')

        if isinstance(args.study_name, type(None)):
            raise ValueError('study name is not defined')

        args.genotype = args.genotype[0]
        args.study_name = args.study_name[0]

        if not os.path.isdir(args.genotype) or not os.path.isdir(args.phenotype) or not os.path.isdir(args.out):
            raise ValueError('{},{} or {} is not a directory'.format(args.genotype, args.phenotype, args.out))

        if not os.path.isdir(os.path.join(args.out, 'encode_genotype')) or not os.path.isdir(
                os.path.join(args.out, 'encode_phenotype')):
            raise ValueError('{} or {} not exist it is up to you to create these directory'.format(
                (os.path.join(args.out, 'encode_genotype')),
                (os.path.join(args.out, 'encode_phenotype'))
            )
            )

        id = np.array(pd.read_hdf(os.path.join(args.genotype, 'individuals', args.study_name + '.h5'),
                                  'individuals').individual.tolist(), dtype=int)

    def single_meta(self, args):

        if len(args.study_name) > 1:
            raise ValueError('Should be only one study name!')

        if len(args.genotype) > 1:
            raise ValueError('Should be only one directory for genotype!')

        if isinstance(args.study_name, type(None)):
            raise ValueError('study name is not defined')

        args.genotype = args.genotype[0]
        args.study_name = args.study_name[0]

        if not os.path.isdir(args.genotype) or not os.path.isdir(args.phenotype) or not os.path.isdir(args.covariates):
            raise ValueError(
                '{},{} or {} is nor a directory'.format(os.path.isdir(args.genotype), os.path.isdir(args.phenotype),
                                                        os.path.isdir(args.covariates)))

    def meta_stage(self, args):

        if isinstance(args.mapper, type(None)):
            raise ValueError(
                'You should define mapper folder --mapper, if it does not exist you should first run Mapper script')

        if isinstance(args.mapper_name, type(None)):
            raise ValueError(
                'You should define mapper name --mapper_name, if it does not exist you should first run Mapper script')

        if len(args.genotype) <= 1:
            raise ValueError('Should be only one directory for genotype!')

        g = [os.path.isdir(i) for i in args.genotype]
        d = [os.path.isdir(i) for i in args.derivatives]

        if np.sum(g) != len(args.genotype) or not os.path.isdir(args.phenotype) or np.sum(d) != len(args.derivatives):
            raise ValueError(
                '{},{} or {} is nor a directory'.format(os.path.isdir(args.genotype), os.path.isdir(args.phenotype),
                                                        os.path.isdir(args.derivatives)))

    def regression(self, args):

        g = [os.path.isdir(i) for i in args.genotype]

        if np.sum(g) != len(args.genotype) or not os.path.isdir(args.phenotype[0]) or not os.path.isdir(
                args.covariates):
            raise ValueError(
                '{},{} or {} is not a directory'.format(os.path.isdir(args.genotype), os.path.isdir(args.phenotype),
                                                        os.path.isdir(args.derivatives)))


########################################################################
########################################################################
########################################################################
class Mapper(object):

    def __init__(self):

        self.name = None
        self.genotype_names = []
        self.dic = OrderedDict()
        self.n_study = 0
        self.values = []  # array (N_study, N_keys), where N_keys = number of ID in reference table
        self.keys = None
        self.n_keys = None
        self.processed = 0
        self._limit = None
        self.reference = False
        self.include = None
        self.exclude = None
        self.include_ind = np.array([])
        self.exclude_ind = np.array([])
        # self.hash=HashTable()
        self.cluster = None
        self.node = None
        self.chunk_pool = None
        self.chunk_size = None
        self.column_names = []
        self.flip = {}  # keys - study names; values - array, with length equal to number of probes
        self.probes = None
        self.encoded = {}

    def chunk_pop(self):
        if self.chunk_pool is None and self.cluster is None:
            raise ValueError('Cluster settings are not defined!')

        if self.chunk_pool is None:
            self.chunk_pool = []
            if self.node is None:
                raise ValueError('cluster setting not defined! (number of nodes / this node number) ')
            self.node = [int(i) for i in self.node]
            N = self.n_keys / self.node[0]
            n = N / self.chunk_size
            if N < 1:
                raise ValueError('Too many nodes! Change chunk size in mapper')
            elif N == 1:
                self.chunk_pool = [[self.node[1] - 1, self.node[1]]]
            else:
                if n == 0:
                    self.chunk_pool = [[N * (self.node[1] - 1), N * (self.node[1] - 1) + N]]
                else:
                    self.chunk_pool = []
                    for i in range(n):
                        self.chunk_pool.append([N * (self.node[1] - 1) + self.chunk_size * i,
                                                N * (self.node[1] - 1) + self.chunk_size * (i + 1)])
                    if (N - n * self.chunk_size) != 0:
                        self.chunk_pool.append([N * (self.node[1] - 1) + self.chunk_size * (n),
                                                N * (self.node[1] - 1) + self.chunk_size * (n) + (
                                                        N - n * self.chunk_size)])

            if self.node[1] == self.node[0]:
                if (self.n_keys - N * self.node[0]) != 0:
                    self.chunk_pool.append([self.chunk_pool[-1][1], self.n_keys])
            self.chunk_pool = self.chunk_pool[::-1]

            print(self.chunk_pool)

        if len(self.chunk_pool) != 0:
            ch = self.chunk_pool.pop()
            print(ch)
            return ch
        else:
            return None

    # @timing
    def fill(self, keys, name, repeats=False, reference=False):  # TODO (middle) remove
        self.reference = reference
        self.reference_name = name
        self.column_names.append(name)
        l = len(keys)
        uniq = np.unique(keys)
        u = len(uniq)
        if not repeats:
            if l != u:
                raise ValueError('length of keys {} does not =  {} number of uniq values!!!'.format(l, u))
        else:
            l = u
            keys = uniq
        self.n_keys = l
        if isinstance(self._limit, type(None)):
            for i, j in enumerate(keys):
                self.dic[j] = [i]
        else:
            for i, j in enumerate(keys):
                if i == self._limit:
                    break
                self.dic[j] = [i]

    # @timing
    def push(self, new_keys, name=None, new_id=True):  # TODO (middle) remove
        if not self.reference and len(self.dic) == 0:
            raise ValueError('You should fill mapper first with ref panel or your own rsids!')
        self.n_study += 1
        if name is not None:
            try:
                self.genotype_names.index(name)
                raise ValueError('Trying to push the same study to mapper')
            except:
                self.genotype_names.append(name)
                self.column_names.append(name)

        if isinstance(self._limit, type(None)):
            for i, j in enumerate(new_keys):
                if self.dic.get(j):
                    self.dic[j] = self.dic[j] + [i]
                else:
                    if new_id:
                        print(('WARNING! You are pushing ID {} which is not present in reference panel!'.format(j)))
                        self.dic[j] = [-1] * (self.n_study + 1)
                        self.dic[j][self.n_study] = i
                    else:
                        continue
        else:
            for i, j in enumerate(new_keys):
                if i == self._limit:
                    break
                if self.dic.get(j):
                    self.dic[j] = self.dic[j] + [i]
                else:
                    self.dic[j] = [-1] * (self.n_study + 1)
                    self.dic[j][self.n_study] = i

        for k in self.dic:
            if len(self.dic[k]) < self.n_study + 1:
                self.dic[k] = self.dic[k] + [-1]

    # @timing
    def load_flip(self, folder, encode=None):

        if folder is None:
            raise ValueError('Mapper is not defined!')
        if not os.path.isdir(folder):
            raise ValueError('{} is not a folder'.format(folder))

        if self.reference_name is None:
            raise ValueError('Reference name for mapper is not defined!')

        if len(glob.glob(folder + 'flip_*')) == 0:
            raise ValueError('There is no flip mapper data in folder {}'.format(folder))

        for j, i in enumerate(self.genotype_names):
            if encode is not None:
                self.encoded[i] = encode[j]
                if encode[j] == 0:
                    self.flip[i] = np.load(os.path.join(folder, 'flip_' + self.reference_name + '_' + i + '.npy'))
                elif encode[j] == 1:
                    flip = np.load(os.path.join(folder, 'flip_' + self.reference_name + '_' + i + '.npy'))
                    self.flip[i] = np.ones(flip.shape[0])
            else:
                self.flip[i] = np.load(os.path.join(folder, 'flip_' + self.reference_name + '_' + i + '.npy'))

    # @timing
    def load(self, folder):
        if folder is None:
            raise ValueError('Mapper is not defined!')
        if not os.path.isdir(folder):
            raise ValueError('{} is not a folder'.format(folder))

        if self.reference_name is None:
            raise ValueError('Reference name for mapper is not defined!')

        if len(glob.glob(folder + 'keys_*')) == 0:
            raise ValueError('There is no mapper data in folder {}'.format(folder))

        keys = glob.glob(folder + 'keys_*')
        if len(keys) > 1:
            raise ValueError('There are more than one reference keys in folder {}'.format(folder))

        self.keys = np.load(os.path.join(keys[0]))  # TODO (middle) not safety to load only one file
        self.n_keys = self.keys.shape[0]

        values = glob.glob(folder + 'values_*')
        if len(values) == 0:
            raise ValueError('There is no mapper data in folder {}'.format(folder))

        if self.genotype_names is None:
            raise ValueError('Genotype names are not defined!')

        self.values = np.zeros((self.n_keys, len(self.genotype_names)))
        for j, i in enumerate(self.genotype_names):
            self.values[:, j] = np.load(os.path.join(folder, 'values_' + self.reference_name + '_' + i + '.npy'))

        if self.n_keys != self.values.shape[0]:
            raise ValueError(
                'Number of indexes {} in mapper values is different from keys length {}!'.format(self.values.shape[0],
                                                                                                 self.n_keys))

        self.n_study = self.values.shape[1]
        if self.n_keys < self.n_study:
            print('WARNING!!! Normally should be more test values that studies! Check your saved data for Mapper')

        print(('You loaded values for {} studies and {} test values'.format(self.n_study, self.n_keys)))

        if self.include is not None or self.exclude is not None:
            self.reference = Reference()
            self.reference.name = self.reference_name
            self.reference.load_index()

    # @timing
    def get(self, chunk_number=None):

        if isinstance(self.keys, type(None)) and isinstance(self.values, type(None)):
            raise ValueError('mapper data is not loaded!')

        if isinstance(self.chunk_size, type(None)):
            raise ValueError('chunk_size and should be defined')

        if self.processed == self.n_keys:
            return None, None

        if chunk_number is not None:
            start = chunk_number[0]
            finish = chunk_number[1]
        else:
            start = self.processed
            finish = self.processed + self.chunk_size if (
                                                                 self.processed + self.chunk_size) < self.n_keys else self.n_keys
            self.processed = finish

        with Timer() as t_q:
            if self.include is not None:
                if len(self.include_ind) == 0:
                    if 'ID' in self.include.columns:
                        self.include_ind = np.in1d(self.keys, self.include.ID)
                        self.include_ind = np.where(self.include_ind == True)[0]
                    # self.query_include=self.reference.index.select('reference',where='ID=self.include.ID')
                    # self.include_ind=self.query_include.index
                    elif 'CHR' in self.include.columns and 'bp' in self.include.columns:
                        self.query_include = self.reference.index.select('reference',
                                                                         where='CHR=self.include.CHR & bp=self.include.bp')
                        self.include_ind = self.query_include.index

            if self.exclude is not None:
                if len(self.exclude_ind) == 0:
                    if 'ID' in self.exclude.columns:
                        self.exclude_ind = np.in1d(self.keys, self.exclude.ID)
                        self.exclude_ind = np.where(self.exclude_ind == True)[0]
                    # self.query_exclude=self.reference.index.select('reference',where='ID=self.exclude.ID')
                    # self.exclude_ind=self.query_exclude.index
                    elif 'CHR' in self.exclude.columns and 'bp' in self.exclude.columns:
                        self.query_exclude = self.reference.index.select('reference',
                                                                         where='CHR=self.exclude.CHR & bp=self.exclude.bp')
                        self.exclude_ind = self.query_exclude.index

        print("Time to select SNPs {}s".format(t_q.secs))

        if self.exclude is not None or self.include is not None:
            self.include_ind = np.setxor1d(self.include_ind, self.exclude_ind)
            if len(self.include_ind) == 0:
                print('None of included ID found in genotype data!')
                return None, None
            else:
                ind = np.intersect1d(np.arange(start, finish), self.include_ind)
        else:
            ind = np.arange(start, finish)
            self.include_ind = np.arange(len(self.values))

        ind = ind.astype('int')
        indexes = self.values[ind, :]
        r = (indexes == -1).any(axis=1)
        ind = ind[~r]
        indexes = indexes[~r]
        keys = self.keys[ind]

        while len(ind) < self.chunk_size:
            if self.processed == self.n_keys:
                break
            if chunk_number is None:
                start = self.processed
                finish = self.processed + self.chunk_size if (
                                                                     self.processed + self.chunk_size) < self.n_keys else self.n_keys
                self.processed = finish
            else:
                ch = self.chunk_pop()
                if ch is not None:
                    start = ch[0]
                    finish = ch[1]
                else:
                    break
            if len(np.append(ind, np.intersect1d(np.arange(start, finish), self.include_ind))) > 1.5 * self.chunk_size:

                if chunk_number is None:
                    self.processed = self.processed - len(np.arange(start, finish))
                    break
                else:
                    print('Added back chunk {}'.format(ch))
                    self.chunk_pool.append(ch)
                    break
            else:
                ind = np.append(ind, np.intersect1d(np.arange(start, finish), self.include_ind))
                ind = ind.astype('int')
                indexes = self.values[ind, :]
                r = (indexes == -1).any(axis=1)
                ind = ind[~r]
                indexes = indexes[~r]
                keys = self.keys[ind]

        if len(ind) == 0:
            return None, None

        return [indexes[:, i].astype(np.int64) for i in range(self.n_study)], keys

    def get_all(self, name, nonempty=True):
        ind = self.genotype_names.index(name)
        indexes = self.values[:, ind]
        if nonempty:
            r = (self.values == -1).any(axis=1)
            if len(r) == self.values.shape[0]:
                raise ValueError('There is no common names between studies')
            indexes = indexes[~r]
        return indexes


########################################################################
########################################################################
########################################################################

class Reference(object):

    def __init__(self):
        self.name = None
        self.path_default = os.path.join(os.environ['HASEDIR'], 'data')
        self.path = {
            '1000Gp1v3_ref': {'table': os.path.join(os.environ['HASEDIR'], 'data', '1000Gp1v3.ref.gz'),
                              'index': os.path.join(os.environ['HASEDIR'], 'data', '1000Gp1v3.ref_info.h5')}
        }
        self.dataframe = None
        self.loaded = False
        self.chunk = 10000
        self.read = 0
        self.columns = ['ID', 'allele1', 'allele2', 'CHR', 'bp']
        self.index = None

    def load(self):
        if self.name is not None:
            if self.path.get(self.name) is not None:
                try:
                    self.dataframe = pd.read_csv(self.path[self.name]['table'], compression='gzip', sep=' ',
                                                 chunksize=self.chunk)
                except:
                    self.dataframe = pd.read_csv(self.path[self.name]['table'], sep=' ', chunksize=self.chunk)
                self.loaded = True
            else:
                if os.path.isfile(os.path.join(self.path_default, self.name)):
                    try:
                        self.dataframe = pd.read_csv(os.path.join(self.path_default, self.name), compression='gzip',
                                                     sep=' ', chunksize=self.chunk)
                    except:
                        self.dataframe = pd.read_csv(os.path.join(self.path_default, self.name), sep=' ',
                                                     chunksize=self.chunk)
                    self.loaded = True
                    self.name = os.path.basename(self.name)
                else:
                    raise ValueError('Unknown reference {}!'.format((self.name)))
        else:
            raise ValueError('Reference name is not define!')

    def load_index(self):
        try:
            self.index = pd.HDFStore(os.path.join(self.path[self.name]['index']), 'r')
        except:
            if os.path.isfile(os.path.join(self.path_default, self.name)):
                self.index = pd.HDFStore(os.path.join(self.path_default, self.name), 'r')
            else:
                raise ValueError('There is {} no index file {}'.format(self.path_default, self.name))

    def __next__(self):
        df = self.dataframe.get_chunk()
        self.read += df.shape[0]
        return df


########################################################################
########################################################################
########################################################################

def study_indexes(args=None, genotype=None, phenotype=None, covariates=None):
    def _get_id(notype):
        if isinstance(notype, type(None)):
            id_p = np.array([], dtype=np.str)
        else:
            if isinstance(notype, tuple):
                id_p = np.array([], dtype=np.str)
                for i in notype:
                    if isinstance(i, dict):
                        id_p = np.append(id_p, i['id'])
                    else:
                        id_p = np.append(id_p, i.get_id())
            else:
                id_p = notype.get_id()
        return id_p.astype(np.str)

    id_c = _get_id(covariates)
    id_p = _get_id(phenotype)
    id_g = _get_id(genotype)

    index_g = np.array([])
    index_p = np.array([])
    index_c = np.array([])

    if args is not None:
        if not isinstance(args.ind_id_inc, type(None)) or not isinstance(args.ind_id_exc, type(None)):
            pass
        if not isinstance(args.cov_name_inc, type(None)) or not isinstance(args.cov_name_exc, type(None)):
            pass

    if len(id_g) != 0 and len(id_p) != 0 and len(id_c) != 0:
        common_id = (np.intersect1d(np.intersect1d(id_g, id_p), id_c))

        index_g = np.array([np.where(id_g == i)[0][0] for i in common_id])
        index_p = np.array([np.where(id_p == i)[0][0] for i in common_id])
        index_c = np.array([np.where(id_c == i)[0][0] for i in common_id])

    elif len(id_g) == 0:
        common_id = (np.intersect1d(id_p, id_c))
        index_p = np.array([np.where(id_p == i)[0][0] for i in common_id])
        index_c = np.array([np.where(id_c == i)[0][0] for i in common_id])

    elif len(id_c) == 0:
        common_id = (np.intersect1d(id_p, id_g))
        index_g = np.array([np.where(id_g == i)[0][0] for i in common_id])
        index_p = np.array([np.where(id_p == i)[0][0] for i in common_id])

    else:
        common_id = (np.intersect1d(id_g, id_c))
        index_g = np.array([np.where(id_g == i)[0][0] for i in common_id])
        index_c = np.array([np.where(id_c == i)[0][0] for i in common_id])

    print(('There are {} common ids'.format(len(common_id))))
    np.savetxt(os.path.join(os.environ['HASEOUT'], 'study_common_id.txt'), common_id, fmt='%s')
    np.savetxt(os.path.join(os.environ['HASEOUT'], 'gen_id.txt'), index_g, fmt='%s')
    np.savetxt(os.path.join(os.environ['HASEOUT'], 'phen_id.txt'), index_p, fmt='%s')
    np.savetxt(os.path.join(os.environ['HASEOUT'], 'cov_id.txt'), index_c, fmt='%s')
    if len(common_id) == 0:
        exit(0)

    return [index_g, index_p, index_c], np.array(common_id)


# @timing
def merge_genotype(genotype, SNPs_index, mapper, flip_flag=True):
    if SNPs_index is None:
        gen = genotype[0].get_next()
        if gen is not None:
            for i in range(1, len(genotype)):
                gen = np.hstack((gen, genotype[i].get_next()))
        return gen
    else:
        if len(genotype) != len(SNPs_index):
            raise ValueError('There are not equal number of genotypes and SNPs indexes {}!={}'.format(len(genotype),
                                                                                                      len(SNPs_index)))
        gen = np.zeros((len(SNPs_index[0]), np.sum([i.folder._data.id.shape[0] for i in genotype])))
        a, b = 0, genotype[0].folder._data.id.shape[0]
        gen[:, a:b] = genotype[0].get(SNPs_index[0], impute=mapper.encoded.get(genotype[0].folder.name, 0))
        a = b
        print(gen.shape)
        if flip_flag:
            if not mapper.encoded.get(genotype[0].folder.name, 0):
                flip = mapper.flip[genotype[0].folder.name][SNPs_index[0]]
                flip_index = (flip == -1)
                gen = np.apply_along_axis(lambda x: flip * (x - 2 * flip_index), 0, gen)
        for i in range(1, len(genotype)):
            g = genotype[i].get(SNPs_index[i], impute=mapper.encoded.get(genotype[i].folder.name, 0))
            print(g.shape)
            b += g.shape[1]
            if flip_flag:
                if not mapper.encoded.get(genotype[i].folder.name, 0):
                    flip = mapper.flip[genotype[i].folder.name][SNPs_index[i]]
                    flip_index = (flip == -1)
                    g = np.apply_along_axis(lambda x: flip * (x - 2 * flip_index), 0, g)

            gen[:, a:b] = g
            a = b

        return gen


def check_converter(converted_folder, study_name):
    l_ind = os.listdir(os.path.join(converted_folder, 'individuals'))
    l_probes = os.listdir(os.path.join(converted_folder, 'probes'))
    l_genotype = os.listdir(os.path.join(converted_folder, 'genotype'))

    if os.path.isdir(os.path.join(converted_folder, 'tmp_files')):
        l_tmp = os.listdir(os.path.join(converted_folder, 'tmp_files'))
        if len(l_tmp) == 0:
            print('There is no tmp files!')
        else:
            try:
                df_tmp = pd.read_csv(os.path.join(converted_folder, 'tmp_files', 'info.txt'), header=None)
            except:
                df_tmp = None

    else:
        df_tmp = None

    if len(l_ind) != 1 or '.h5' not in l_ind[0]:
        print('There should be 1 hdf5 file in individuals folder!')

    else:
        df_ind = pd.read_hdf(os.path.join(converted_folder, 'individuals', study_name + '.h5'), 'individuals')
        if df_tmp is not None and int(df_tmp[0][0]) != df_ind.shape[0]:
            print(('Converted number of IDs {} != {} original number'.format(df_ind.shape[0], df_tmp[0][0])))
        else:
            print(('Number of individuals {} '.format(df_ind.shape[0])))
            print(df_ind.head())

    if len(l_probes) != 2:
        print(('There should be 2 files in probes folder, you have {}!'.format(len(l_probes))))
        probes = pd.HDFStore(os.path.join(converted_folder, 'probes', study_name + '.h5'), 'r')
        N_probs = probes.get_storer('probes').nrows

    else:
        probes = pd.HDFStore(os.path.join(converted_folder, 'probes', study_name + '.h5'), 'r')
        N_probs = probes.get_storer('probes').nrows

    if df_tmp is not None and int(df_tmp[0][1]) != N_probs:
        print(("Converted number of variants {} diff from  original number {}".format(N_probs, df_tmp[0][1])))

    else:
        print(('Converted number of variants {}'.format(N_probs)))

        print(probes.select('probes', start=0, stop=10))
        probes.close()

    if len(l_genotype) == 0:
        print('There is no genotype data converted!')
    else:
        index = 0
        for i in l_genotype:
            n, m = h5py.File(os.path.join(converted_folder, 'genotype', i), 'r')['genotype'].shape
            index += n

        if index != N_probs:
            print(('Number of variants in genotype folder {} != {} number in probes file'.format(index, N_probs)))

        else:
            print(('Number of variants in genotype folder {}'.format(index)))

        print(h5py.File(os.path.join(converted_folder, 'genotype', l_genotype[0]), 'r')['genotype'][...])


if __name__ == '__main__':
    print('tools')



import argparse
import numpy as np
import pandas as pd
import os
import argparse
import numpy as np
import pandas as pd
import os
from scipy import sparse


def main():
    """
    args: 
    snp: the name of the column in the topology.csv dataset with the ID for the SNP column
    gene: the name of the column in the topology.csv dataset with the ID for the gene column
    direc: (Optional) the directory where the topology.csv file is located, if omitted it takes the current directory
    file_name: (Optional) the name of the file to save as, defaults to "SNP_gene_mask"

    Return: SNP_gene_mask.npz, the .npz file corresponding to the topology.csv
    """
    parser = argparse.ArgumentParser(description="A simple script with command-line arguments")
    parser.add_argument("--snp", help="Your snp", required=True)
    parser.add_argument("--gene", help="Your gene", required=True)
    parser.add_argument("--direc", help="Your Directory", required=False)
    parser.add_argument("--file_name", help="Your file name", default="SNP_gene_mask", required=False)
    args = parser.parse_args()
    
    if args.direc:
        try:
            os.chdir(args.direc)
            print(f"Navigated to directory: {os.getcwd()}")
        except FileNotFoundError:
            print(f"Directory '{args.direc}' not found.")
    
    snp_level = args.snp
    gene_level = args.gene
    topology = pd.read_csv("topology.csv")
    data = np.ones(len(topology), np.bool)
    coord = (topology[snp_level].values, topology[gene_level].values)
    SNP_gene_matrix = sparse.coo_matrix(((data), coord), shape=(topology[snp_level].max()+1, topology[gene_level].max()+1))
    file_name = args.file_name
    sparse.save_npz(file_name, SNP_gene_matrix)

if __name__ == "__main__":
    main()





import argparse
import os

import numpy as np
import pandas as pd

parser = argparse.ArgumentParser(description='Script to edit phenotypes ids for HASE')
parser.add_argument("-i", required=True, type=str, help="path to nparrays of phenotypes")
parser.add_argument("-ids", type=str, help="path to one column data frame without headers"
                                           "with a list of ids to exclude from analysis")
parser.add_argument("-out", required=True, type=str, help="path to save new phenotypes")

args = parser.parse_args()
print(args)

df_path = args.ids
pheno_path = args.i
out_path = args.out

if not os.path.isdir(out_path):
    print("Creating directory {}".format(out_path))
    os.makedirs(out_path)
else:
    if len(os.listdir(out_path)) != 0:
        raise ValueError('Output folder is not empty!')

print("Crating soft link for original pheno files in {}".format(out_path))

for i in os.listdir(pheno_path):
    if 'info_dic' not in i:
        os.symlink(os.path.join(pheno_path, i), os.path.join(out_path, i))

info_dic = np.load(os.path.join(pheno_path, 'info_dic.npy')).item()

df = pd.read_csv(df_path, header=None, sep=',')
df.columns = ['ID']
df['ID'] = df.ID.astype('str')

index = pd.Series(info_dic['id']).isin(df.ID)
index = np.where(index == True)[0]

if len(index) != 0:
    print('There are {} ids to remove!'.format(len(index)))
    for i in index:
        info_dic['id'][i] = "remove_" + info_dic['id'][i]
else:
    print('No ids to remove!')

np.save(os.path.join(out_path, 'info_dic.npy'), info_dic)



import seaborn as sns
import os
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
import sklearn.metrics as skm
import scipy
import sys
import pandas as pd
import numpy as np
    
# import numba as nb
import itertools

from scipy.sparse import coo_matrix
            
            
def GenNet_pairwise_interactions_topn(w_input, w_later, mask, n):
    '''
    NID fast implementation for topn with:
    
    n: number of interactions per gene
    n cannot be a larger number than the smallest gene size
    '''

    mask_row = np.array(mask.row)
    mask_col = np.array(mask.col)
    mask_data = abs(np.array(w_input))
    
    n_genes = int(mask.shape[1])

    if n=="auto":
       n=10e6
    
    if (min(mask.sum(axis=0)) < n).any():  # Cannot get the top n if genes have less than n
        n = np.min(mask.sum(axis=0))

    num_combinations = int(np.round(np.math.factorial(n) / (np.math.factorial(2) * np.math.factorial((n - 2))) ))

    SNP_coord, strength = get_p_interactions(w_input, w_later, num_combinations, mask_row, mask_col, mask_data, n_genes, n)

    strength, ix = min_combination_rows(strength)  # take the min values of all the combinations
    strength = np.multiply(strength, w_later.flatten())  # multiply those with the w_rest
    
    
    # matrix of the combinations of SNP coord                                                                    
    df1 = pd.DataFrame(SNP_coord.astype(int).astype(str))  
    df1 = pd.DataFrame(df1.values + " ")      
    df1 = pd.DataFrame(df1.values[ix].sum(axis=1))
    SNP_coord = df1.values

    ind = np.argsort(strength, axis=0)[::-1]              # order these by value, descending so inverse
    strength = np.take_along_axis(strength, ind, axis=0)                  # apply ordering strength
    SNP_coord = np.take_along_axis(SNP_coord, ind, axis=0)  

    interaction_ranking = pd.DataFrame(np.array([SNP_coord.flatten(), strength.flatten()]).T, columns=["coord","strength"])
    interaction_ranking = interaction_ranking.sort_values("strength", ascending=False)

    return interaction_ranking  


def get_p_interactions(w_input, w_later,num_combinations, mask_row, mask_col, mask_data, n_genes, n):
    SNP_coord = np.zeros(shape=(n, n_genes)) # all snp numbers, col is gene number
    strength = np.zeros(shape=(n , n_genes)) # 1 value for each SNPcoord
    for gene_id in range(n_genes): # nb.range()  # over all genes
        top_coord = np.argsort(-mask_data[mask_col == gene_id])[:n] # select the highest n values of the gene
        SNP_coord[:, gene_id] = mask_row[mask_col == gene_id][top_coord]    #  select the coordinate of the highest value
        strength[:,gene_id]  = mask_data[mask_col == gene_id][top_coord]    #  select the weights ofthe highest value
    return SNP_coord, strength

def min_combination_rows(comb_array):
    n = comb_array.shape[0]
    ix = np.indices((n,n))[:, ~np.tri(n, k=0, dtype=bool)].T
    minimum_array = comb_array[ix].min(axis=1)
    return minimum_array, ix 

def Get_weight_tsang(model, layer_n, masks):
    '''
    Get the proper w_in and w_out for the NID algorithm.
    
    For multiple layers:
    w_in, w_out = Get_weight_tsang(model, 2, genemask)
    print(w_in.shape, w_out.shape)
    
    print('causal_snps', causal_snps)
    interaction_ranking1 = GenNet_pairwise_interactions_topn(w_in[:,1] ,w_out[:,1], genemask, n=4)
    interaction_ranking2 = GenNet_pairwise_interactions_topn(w_in[:,0] ,w_out[:,0], genemask, n=4)

    interaction_ranking = interaction_ranking1.append(interaction_ranking2)
    interaction_ranking = interaction_ranking.sort_values("strength", ascending =False)
    interaction_ranking.head(20)
    '''
    w_in_layer = model.layers[layer_n]
    w_in = abs(w_in_layer.get_weights()[0])
    print("From", w_in_layer.__class__.__name__, w_in.shape)
    mask_count = 0   
    w_out = 0

    for i in range(len(model.layers)-1, -1, -1):
        if layer_n == i:
            return abs(w_in), abs(w_out)

        layer = model.layers[i]
        layer_name = model.layers[i].name
        layer_type = model.layers[i].__class__.__name__
        print(i, layer_type)

        if "LocallyDirected1D" in layer_type:
            mask = masks[mask_count]
            weights = layer.get_weights()[0]
            w_filter = []

            for j in range(weights.shape[1]):
                mask.values = abs(weights[:,j])
                w_filter.append(np.matmul(mask.todense(), w_out)) 

            w_out = np.concatenate(w_filter, axis=-1) 


            names = ['weights_'+str(i)+ "_"+str(j) for j in range(weights.ndim)]

            mask_count +=1 

        elif "LocallyConnected1D" in layer_type:
            weights = layer.get_weights()[0]
            w_filter = []
            if weights.ndim == 3:
                for j in range(weights.shape[1]):
                    w_filter.append(w_out * weights[:,j])  
                w_out = np.concatenate(w_filter, axis=-1) 
            else:
                print("unexpected shape", weights.shape)

        elif "Dense" in layer_type and "_cov" not in layer_name:
            weights = abs(layer.get_weights()[0])
            if w_out == 0:
                w_out = abs(weights)   
            else:
                w_out = np.matmul(w_out, np.abs(weights))
        else:
            print(" -- Skipping", layer_type, '--')
    
    return w_in, w_out 


import itertools
def GenNet_pairwise_interactions_simplified_mask(w_input, w_later, mask):
    """first simple implementation"""
    mask = mask * 1
    interaction_ranking = []
    list_of_combinations = []
    for gene_id in range(mask.shape[1]):
        neuron_combinations=list(itertools.combinations(mask.row[mask.col == gene_id], 2 ))

    for candidate in neuron_combinations:
        strength = (np.minimum(w_input[candidate[0]], w_input[candidate[1]])*w_later[gene_id]).sum()
        interaction_ranking.append(((candidate[0], candidate[1]), strength))

    interaction_ranking.sort(key=lambda x: x[1], reverse=True)
    return interaction_ranking





    
    
            

    


import statsmodels.api as sm
import tqdm
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
import tensorflow as tf
import os
import tensorflow.keras as K
import scipy
import sys
import argparse
import time
import tqdm
import scipy
import scipy.sparse

from collections import defaultdict

class RLIPP:
    def __init__(self, model, xtrain, ytrain, child_layer=None, masks=None, input_regression_path='/'):
        self.model = model
        self.xtrain = xtrain
        self.ytrain = ytrain
        self.child_layer = child_layer
        self.masks = masks
        self.input_regression_path = input_regression_path
        self.layer_names = None
        self.pd_RLIPP = pd.DataFrame([], columns= ['layer_name','layer', 'node', 'RLIPP'])

    def calculate_modelwise(self):
        if self.child_layer is None:
            self.layer_names = [x.name for x in self.model.layers if "activation" in x.name]
            assert len([x.__class__.__name__ for x in self.model.layers if "LocallyDirected" in x.__class__.__name__]) == len(self.masks)
        
        self.model  = self.get_node_correlation()
        self.model  = self.get_input_correlation()

        for layer_num, layer in enumerate(self.model.layers):
            layer_type = self.model.layers[layer_num].__class__.__name__ 

            if ("activation" in layer.name) and ("snp" not in layer.name):
                self.RLIPP(layer_num)
            elif "LocallyDirected1D" in layer_type:
                self.get_RLIPP_LD
            elif "LocallyConnected1D" in layer_type :
                self.get_RLIPP_LC
            elif "Dense" in layer_type :
                self.get_RLIPP_dense
            else:
                print(" -- Skipping", layer_type, '--') 

        return self.pd_RLIPP
    
    def get_input_correlation(self):
        assert self.xtrain.ndim == 3, "xtrain needs to be 3 dimensional with (npat, nfeatures, 3 or 1)"
        prsqrt_input = np.full((self.xtrain.shape[1], self.xtrain.shape[2]), np.NaN)

        for SNP_num in tqdm.tqdm(range(self.xtrain.shape[1])):
            for allele in range(self.xtrain.shape[2]):
                try:
                    regresult = sm.Logit(self.ytrain, self.xtrain[:,SNP_num, allele]).fit(disp=0)
                    prsqrt_input[SNP_num, allele] = regresult.prsquared
                except:
                    prsqrt_input[SNP_num, allele] = np.NaN

        self.model.layers[0].prsqrt = prsqrt_input
        return self.model
 
    
    def get_node_correlation(self):
        
        for layer_num, layer in enumerate(self.model.layers):
            if 'activation' in layer.name:

                intermediate_layer_model = tf.keras.Model(inputs=self.model.input,
                                                 outputs=self.model.get_layer(layer.name).output)
                predictions = intermediate_layer_model.predict(self.xtrain)
                predictions = np.reshape(predictions, (self.xtrain.shape[0], -1))
                prsqrt_node = np.full((predictions.shape[1]), np.NaN)

                for node in range(predictions.shape[1]):
                    if len(np.unique(predictions[:,node])) > 1:
                        regresult = sm.Logit(self.ytrain, predictions[:,node]).fit(disp=0)
                        prsqrt_node[node] = regresult.prsquared
                    else:
                        prsqrt_node[node] = np.NaN
            else:
                prsqrt_node = np.full((0), np.NaN)

            self.layers[layer_num].prsqrt = prsqrt_node
            
            
    def get_RLIPP(self, parent_prsqrt, child_prsqrt, node_num, layer_num):
        RLIPP_score = float((parent_prsqrt - child_prsqrt) / child_prsqrt)

        self.pd_RLIPP.append({"layer":layer_num, 
                                "layer_name":self.model.layers[layer_num].name,
                                "node":node_num,
                                "RLIPP":RLIPP_score}, ignore_index=True)    
         
            
    def get_RLIPP_LC(self, layer_num,):
        parent_prsqrt = self.model.layers[layer_num].prsqrt
        childeren_prsqrt = self.get_child_nodes(self.model, layer_num)

        kernelsize = int(len(childeren_prsqrt) / len(parent_prsqrt))

        childeren_prsqrt = np.reshape(childeren_prsqrt, (-1, kernelsize))

        # calculate the RLIPP for the right parent-child
        for node_num, child_prsqrts in enumerate(childeren_prsqrt):
            for child_prsqrt in child_prsqrts:
                  self.get_RLIPP(parent_prsqrt[node_num], child_prsqrt, node_num, layer_num)


    def get_RLIPP_LD(self, layer_num):
        parent_prsqrt = self.model.layers[layer_num].prsqrt
        childeren_prsqrt = self.get_child_nodes(self.model, layer_num)

        num_mask = -1
        for i in range(layer_num):
            if "LocallyDirected" in self.model.layers[i].__class__.__name__:
                num_mask +=1

        mask = self.masks[num_mask]

        nfilters = int(len(parent_prsqrt) / mask.shape[1])
        parent_prsqrt = np.reshape(parent_prsqrt, (mask.shape[1], nfilters))
        
        for cfilter in range(nfilters):                       # per filter
            cparent_prsqrt =  parent_prsqrt[:,cfilter]        # select the parent node
            for node_num in range(mask.shape[1]):            # per gene
                children = mask.row[mask.col==node_num]      # select children
                for child_prsqrt in childeren_prsqrt[children]:  # per child
                    self.get_RLIPP(cparent_prsqrt[node_num], child_prsqrt, node_num, layer_num)    




    def get_RLIPP_dense(self, layer_num):
        parent_prsqrt = self.model.layers[layer_num].prsqrt
        childeren_prsqrt = self.get_child_nodes(self.model, layer_num)

        # calculate the RLIPP for the right parent-child
        for node_num, child_prsqrt in enumerate(childeren_prsqrt):
            self.get_RLIPP(parent_prsqrt, child_prsqrt, node_num, layer_num)




import sys
import json
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import sklearn.metrics as skm
import tensorflow as tf
import tensorflow.keras as K
from argparse import Namespace
tf.keras.backend.set_epsilon(0.0000001)
import os
import seaborn as sns
from sklearn.metrics import mean_squared_error, explained_variance_score, r2_score


def get_SLURM_id():
    SlURM_JOB_ID = "unknown"
    try:
        print('SlURM_JOB_ID',os.environ["SLURM_JOB_ID"])
        SlURM_JOB_ID = os.environ["SLURM_JOB_ID"]
    except:
        print("no slurm id")
    return SlURM_JOB_ID


def use_mixed_precision():
    from tensorflow.keras import mixed_precision
    policy = mixed_precision.Policy('mixed_float16')
    mixed_precision.set_global_policy(policy)

def get_paths(args):
    folder = ("GenNet_experiment_" + str(args.ID))

    if args.out == "undefined":
        resultpath = os.getcwd() + "/results/" + folder + "_" + str(args.suffix) + "/"
    else:
        resultpath = str(args.out) + "/" + folder + "_" + str(args.suffix) + "/"

    if not os.path.exists(resultpath):
        print("Resultspath did not exist but is made now")
        os.makedirs(resultpath)

    return folder, resultpath


def explode(df, cols, split_on=','):
    """
    Explode dataframe on the given column, split on given delimeter
    """
    cols_sep = list(set(df.columns) - set(cols))
    df_cols = df[cols_sep]
    explode_len = df[cols[0]].str.split(split_on).map(len)
    repeat_list = []
    for r, e in zip(df_cols.as_matrix(), explode_len):
        repeat_list.extend([list(r)] * e)
    df_repeat = pd.DataFrame(repeat_list, columns=cols_sep)
    df_explode = pd.concat([df[col].str.split(split_on, expand=True).stack().str.strip().reset_index(drop=True)
                            for col in cols], axis=1)
    df_explode.columns = cols
    return pd.concat((df_repeat, df_explode), axis=1)


def sensitivity(y_true, y_pred):
    y_true = tf.cast(y_true, tf.float32)
    y_pred = tf.cast(y_pred, tf.float32)
    y_pred = tf.keras.backend.clip(y_pred, 0, 1)
    y_true = tf.keras.backend.clip(y_true, 0, 1)

    y_pred = tf.keras.backend.round(y_pred)

    true_p = K.backend.sum(K.backend.round(y_pred) * y_true)
    pos = tf.keras.backend.sum(y_true)
    sensitivity = tf.keras.backend.clip((true_p / (pos + 0.00001)), 0, 1)
    return sensitivity


def specificity(y_true, y_pred):
    y_true = tf.cast(y_true, tf.float32)
    y_pred = tf.cast(y_pred, tf.float32)
    y_pred = tf.keras.backend.clip(y_pred, 0, 1)
    y_true = tf.keras.backend.clip(y_true, 0, 1)

    neg_y_true = 1 - y_true
    neg_y_pred = 1 - K.backend.round(y_pred)
    fp = K.backend.sum(neg_y_true * K.backend.round(y_pred))
    tn = K.backend.sum(neg_y_true * neg_y_pred)
    specificity = tn / (tn + fp + 0.00001)
    return tf.keras.backend.clip(specificity, 0, 1)


def binary_accuracy(y_true, y_pred):
    return K.backend.mean(K.backend.equal(y_true, K.backend.round(y_pred)))


def evaluate_performance_regression(y, p):
    y = y.flatten()
    p = p.flatten()
    explained_variance = explained_variance_score(y, p)
    mse = mean_squared_error(y, p)
    r2 = r2_score(y, p)
    print("Mean squared error =", mse)
    print("Explained variance =", explained_variance)
    # print("maximum error =", maximum_error)
    print("r2 =", r2)

    plt.figure()
    df = pd.DataFrame([])
    df["truth"] = y
    df["predicted"] = p

    fig = sns.jointplot(x="truth", y="predicted", data=df, alpha=0.5)
    return fig, mse, explained_variance, r2


def evaluate_performance_classification(y, p):
    print("\n")
    print("Confusion matrix")
    confusion_matrix = skm.confusion_matrix(y, p.round())
    print(confusion_matrix)

    fpr, tpr, thresholds = skm.roc_curve(y, p)
    roc_auc = skm.auc(fpr, tpr)
    print("\n")
    print("Area under the Curve (AUC) = ", roc_auc)

    specificity = confusion_matrix[0, 0] / (confusion_matrix[0, 0] + confusion_matrix[0, 1])
    print('Specificity = ', specificity)

    sensitivity = confusion_matrix[1, 1] / (confusion_matrix[1, 0] + confusion_matrix[1, 1])
    print('Sensitivity = ', sensitivity)
    print("F_1 score = " + str(skm.f1_score(y, p.round())))
    plt.figure()
    plt.plot(fpr, tpr, color='darkorange',
             lw=2, label='ROC curve (area = %0.2f)' % roc_auc)
    plt.plot([0, 1], [0, 1], color='navy', lw=2, linestyle='--')
    plt.plot(1 - specificity, sensitivity, color='b', marker='o')
    plt.xlim([0.0, 1.0])
    plt.ylim([0.0, 1.05])
    plt.xlabel('False Positive Rate')
    plt.ylabel('True Positive Rate')
    plt.title('Receiver operating characteristic example')
    plt.legend(loc="lower right")

    return roc_auc, confusion_matrix


def create_importance_csv(datapath, model, masks):
    network_csv = pd.read_csv(datapath + "/topology.csv")

    coordinate_list = []
    for i, mask in zip(np.arange(len(masks)), masks):
        coordinates = pd.DataFrame([])

        if (i == 0):
            if 'chr' in network_csv.columns:
                coordinates["chr"] = network_csv["chr"]
        coordinates["node_layer_" + str(i)] = mask.row
        coordinates["node_layer_" + str(i + 1)] = mask.col
        coordinates = coordinates.sort_values("node_layer_" + str(i), ascending=True)
        coordinates["weights_" + str(i)] = model.get_layer(name="LocallyDirected_" + str(i)).get_weights()[0]

        coordinate_names = network_csv[["layer" + str(i) + "_node", "layer" + str(i) + "_name"]].drop_duplicates()
        coordinate_names = coordinate_names.rename({"layer" + str(i) + "_node": "node_layer_" + str(i)}, axis=1)
        coordinates = coordinates.merge(coordinate_names, on="node_layer_" + str(i))
        coordinate_list.append(coordinates)

        if i == 0:
            total_list = coordinate_list[i]
        else:
            total_list = total_list.merge(coordinate_list[i], on="node_layer_" + str(i))

    i += 1
    coordinates = pd.DataFrame([])
    coordinates["weights_" + str(i)] = model.get_layer(name="output_layer").get_weights()[0].flatten()
    coordinates["node_layer_" + str(i)] = np.arange(len(coordinates))
    coordinate_names = network_csv[["layer" + str(i) + "_node", "layer" + str(i) + "_name"]].drop_duplicates()
    coordinate_names = coordinate_names.rename({"layer" + str(i) + "_node": "node_layer_" + str(i)}, axis=1)
    coordinates = coordinates.merge(coordinate_names, on="node_layer_" + str(i))
    total_list = total_list.merge(coordinates, on="node_layer_" + str(i))
    total_list["raw_importance"] = total_list.filter(like="weights").prod(axis=1)
    return total_list


def query_yes_no(question, default="yes"):
    """Ask a yes/no question via raw_input() and return their answer.

    "question" is a string that is presented to the user.
    "default" is the presumed answer if the user just hits <Enter>.
        It must be "yes" (the default), "no" or None (meaning
        an answer is required of the user).

    The "answer" return value is True for "yes" or False for "no".
    """
    valid = {"yes": True, "y": True, "ye": True,
             "no": False, "n": False}
    if default is None:
        prompt = " [y/n] "
    elif default == "yes":
        prompt = " [Y/n] "
    elif default == "no":
        prompt = " [y/N] "
    else:
        raise ValueError("invalid default answer: '%s'" % default)

    while True:
        sys.stdout.write(question + prompt)
        choice = input().lower()
        if default is not None and choice == '':
            return valid[default]
        elif choice in valid:
            return valid[choice]
        else:
            sys.stdout.write("Please respond with 'yes' or 'no' "
                             "(or 'y' or 'n').\n")



def save_train_arguments(args, filename="train_args.json"):
    """
    Save the training arguments to a JSON file.

    Parameters:
    - args: The arguments object (typically an instance of ArgumentParser or similar).
    - filename: The name of the file where the arguments will be saved.
    """

    # Convert args to a dictionary, taking care of non-serializable types if necessary
    print("saving arguments")
    args_dict = vars(args)  # Convert the Namespace to a dictionary
    with open(args.resultpath + "/" + filename, 'w') as file:
        json.dump(args_dict, file, cls=NumpyEncoder, indent=4)



def load_train_arguments(args, filename="train_args.json"):
    """
    Load the training arguments from a JSON file and update missing arguments.

    Parameters:
    - filename: The name of the file from which the arguments will be loaded.

    Returns:
    - args: The updated arguments object.
    """
    with open(args.resultpath + "/" + filename, 'r') as file:
        args_dict = json.load(file)
    
    # Update only the missing attributes in args
    for key, value in args_dict.items():
        if not hasattr(args, key):
            setattr(args, key, value)
    
    return args


class NumpyEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, (np.int_, np.intc, np.intp, np.int8,
            np.int16, np.int32, np.int64, np.uint8,
            np.uint16, np.uint32, np.uint64)):
            return int(obj)
        elif isinstance(obj, (np.float_, np.float16, np.float32, 
            np.float64)):
            return float(obj)
        elif isinstance(obj, (np.ndarray,)):  # Handle numpy arrays
            return obj.tolist()
        return super(NumpyEncoder, self).default(obj)



import glob
import os
import sys
import argparse
import h5py
import numpy as np
import pandas as pd
import tables
import tqdm
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
sys.path.insert(1, os.path.join(sys.path[0], '..'))
from gennet.utils.Utility_functions import query_yes_no
from gennet.utils.hase.config import basedir, CONVERTER_SPLIT_SIZE, PYTHON_PATH

os.environ['HASEDIR'] = basedir
if PYTHON_PATH is not None:
    for i in PYTHON_PATH: sys.path.insert(0, i)
from gennet.utils.hase.hdgwas.tools import Timer, check_converter
from gennet.utils.hase.hdgwas.converter import GenotypePLINK, GenotypeMINIMAC, GenotypeVCF
from gennet.utils.hase.hdgwas.data import Reader


def hase_convert(args):
    if (os.path.exists(args.outfolder + '/probes/')) and (os.path.exists(args.outfolder + '/genotype/')) and (
            os.path.exists(args.outfolder + '/individuals/')):
        print("The folders: probes, genotype and individuals already exist. Data seems already in HASE format. Delete "
              "the folders if the files are not converted properly. Continuing with the current files:")
        return
    else:
        print('using', args.outfolder)

    R = Reader('genotype')

    R.start(args.genotype[0], vcf=args.vcf)

    with Timer() as t:
        if R.format == 'PLINK':
            G = GenotypePLINK(args.study_name[0], reader=R)
            G.split_size = CONVERTER_SPLIT_SIZE
            G.plink2hdf5(out=args.out)

        elif R.format == 'MINIMAC':
            G = GenotypeMINIMAC(args.study_name[0], reader=R)
            G.split_size = CONVERTER_SPLIT_SIZE
            G.MACH2hdf5(args.out, id=args.id)

        elif R.format == 'VCF':
            G = GenotypeVCF(args.study_name[0], reader=R)
            G.split_size = CONVERTER_SPLIT_SIZE
            G.VCF2hdf5(args.out)
        else:
            raise ValueError('Genotype data should be in PLINK/MINIMAC/VCF format and alone in folder')

    check_converter(args.out, args.study_name[0])
    args.outfolder = args.genotype
    print(('Time to convert all data: {} sec'.format(t.secs)))
    return


def merge_hdf5_hase(args):
    print(args.genotype, type(args.genotype))
    filepath_hase = args.genotype + '/genotype/{}_' + args.study_name + '.h5'
    g = h5py.File(filepath_hase.format(0), 'r')['genotype']
    num_pat = g.shape[1]
    number_of_files = len(glob.glob(args.genotype + "/genotype/*.h5"))
    print('number of files ', number_of_files)

    f = tables.open_file(args.outfolder + args.study_name + '_step2_merged_genotype.h5', mode='w')
    atom = tables.Int8Col()
    filter_zlib = tables.Filters(complib='zlib', complevel=args.comp_level)
    f.create_earray(f.root, 'data', atom, (0, num_pat), filters=filter_zlib)
    f.close()

    print("\n merge all files...")
    f = tables.open_file(args.outfolder + args.study_name + '_step2_merged_genotype.h5', mode='a')
    for i in tqdm.tqdm(range(number_of_files)):
        gen_tmp = h5py.File(filepath_hase.format(i), 'r')['genotype']
        f.root.data.append(np.array(np.round(gen_tmp[:, :]), dtype=int))
    f.close()

    args.outfolder = args.genotype


def impute_hase_hdf5_no_chunk(args):
    t = tables.open_file(args.genotype + args.study_name + '_step2_merged_genotype.h5', mode='r')
    print('merged shape =', t.root.data.shape)
    num_SNPS = t.root.data.shape[0]
    num_pat = t.root.data.shape[1]

    hdf5_name = args.study_name + '_step3_genotype_no_missing.h5'
    p = pd.read_hdf(args.genotype + '/probes/' + args.study_name + ".h5")
    print('probe shape =', p.shape)

    print("\n impute missing...")
    f = tables.open_file(args.outfolder + args.study_name + '_step3_genotype_no_missing.h5', mode='w')
    atom = tables.Int8Col()

    filter_zlib = tables.Filters(complib='zlib', complevel=args.comp_level)
    f.create_earray(f.root, 'data', atom, (0, num_pat), filters=filter_zlib)
    f.close()

    stdSNPs = np.zeros(num_SNPS)
    f = tables.open_file(args.outfolder + args.study_name + '_step3_genotype_no_missing.h5', mode='a')

    for i in tqdm.tqdm(range(t.root.data.shape[0])):
        d = t.root.data[i, :].astype("float32")
        m = np.where(d == 9)
        d[m] = np.nan
        d[m] = np.nanmean(d)
        d = d[np.newaxis, :]
        f.root.data.append(np.round(d).astype(np.int8))
        stdSNPs[i] = np.std(d)
    f.close()
    t.close()

    np.save(args.outfolder + args.study_name + '_std.npy', stdSNPs)

    args.outfolder = args.genotype

    return hdf5_name


def impute_hase_hdf5(args):
    t = tables.open_file(args.genotype + args.study_name + '_step2_merged_genotype.h5', mode='r')
    print('merged shape =', t.root.data.shape)
    num_SNPS = t.root.data.shape[0]
    num_pat = t.root.data.shape[1]

    hdf5_name = args.study_name + '_step3_genotype_no_missing.h5'
    p = pd.read_hdf(args.genotype + '/probes/' + args.study_name + ".h5")
    print('probe shape =', p.shape)

    print("\n impute missing...")
    f = tables.open_file(args.outfolder + args.study_name + '_step3_genotype_no_missing.h5', mode='w')
    atom = tables.Int8Col()

    filter_zlib = tables.Filters(complib='zlib', complevel=args.comp_level)
    f.create_earray(f.root, 'data', atom, (0, num_pat), filters=filter_zlib)
    f.close()

    stdSNPs = np.zeros(num_SNPS)
    f = tables.open_file(args.outfolder + args.study_name + '_step3_genotype_no_missing.h5', mode='a')

    chunk = args.tcm // num_SNPS
    chunk = int(np.clip(chunk, 1, num_pat))
    print(chunk)

    for part in tqdm.tqdm(range(int(np.ceil(num_SNPS / chunk) + 1))):
        begins = part * chunk
        tills = min(((part + 1) * chunk), num_SNPS)
        d = t.root.data[begins:tills, :].astype("float32")
        d[d == 9] = np.nan
        a = np.where(np.isnan(d), np.ma.array(d, mask=np.isnan(d)).mean(axis=1)[:, np.newaxis], d)
        stdSNPs[begins:tills] = np.std(a, axis=1)
        f.root.data.append(np.round(d).astype(np.int8))
    f.close()
    t.close()

    np.save(args.outfolder + args.study_name + '_std.npy', stdSNPs)
    args.outfolder = args.genotype
    return hdf5_name


def exclude_variants(args):
    print("Selecting the variants..")
    t = tables.open_file(args.genotype + args.study_name + '_step3_genotype_no_missing.h5', mode='r')
    data = t.root.data
    num_pat = data.shape[1]
    num_variants = data.shape[0]

    used_indices = pd.read_csv(args.variants, header=None)

    hdf5_name = args.study_name + '_step4_genotype_selected_variants.h5'

    if len(used_indices) == num_variants:
        used_indices = used_indices.index.values[used_indices.values.flatten()]
        f = tables.open_file(args.outfolder + args.study_name + '_step4_genotype_selected_variants.h5', mode='w')
        f.create_earray(f.root, 'data', tables.IntCol(), (0, num_pat), expectedrows=len(used_indices),
                        filters=tables.Filters(complib='zlib', complevel=args.comp_level))
        f.close()

        f = tables.open_file(args.outfolder + args.study_name + '_step4_genotype_selected_variants.h5', mode='a')
        for feat in tqdm.tqdm(used_indices):
            a = data[feat, :]
            a = np.reshape(a, (1, -1))
            f.root.data.append(a)
        f.close()
        t.close()
        args.outfolder = args.genotype
        return hdf5_name

    else:
        print("Something wrong with the included_snps file.")
        print("Expected " + str(num_variants) + "but got " + str(len(used_indices)))
        print('Used indices looks like:', used_indices)
        exit()


def transpose_genotype(args):
    step4_name = args.genotype + '/' + args.study_name + '_step4_genotype_selected_variants.h5'
    step3_name = args.genotype + '/' + args.study_name + '_step3_genotype_no_missing.h5'
    step2_name = args.genotype + '/' + args.study_name + '_step2_merged_genotype.h5'
    
    if (os.path.exists(step4_name)):
        t = tables.open_file(step4_name, mode='r')
    elif (os.path.exists(step3_name)):
        print('WARNING skipped step 4, all variants are used: using', step3_name)
        t = tables.open_file(step3_name, mode='r')
    elif (os.path.exists(step2_name)):
        print('WARNING skipped step 3, only skip this step if you are sure there are no missing variants (i.e: genotype only has values 0,1,2)', step2_name)
        t = tables.open_file(step2_name, mode='r')
    else: 
        print('no valid genotype found')

    data = t.root.data
    num_pat = data.shape[1]
    num_feat = data.shape[0]
    chunk = args.tcm // num_feat
    chunk = int(np.clip(chunk, 1, num_pat))
    print("chuncksize =", chunk)

    f = tables.open_file(args.outfolder + '/genotype.h5', mode='w')
    f.create_earray(f.root, 'data', tables.IntCol(), (0, num_feat), expectedrows=num_pat,
                    filters=tables.Filters(complib='zlib', complevel=args.comp_level))
    f.close()

    f = tables.open_file(args.outfolder + '/genotype.h5', mode='a')

    for pat in tqdm.tqdm(range(int(np.ceil(num_pat / chunk) + 1))):
        begins = pat * chunk
        tills = min(((pat + 1) * chunk), num_pat)
        a = np.array(data[:, begins:tills], dtype=np.int8)
        a = a.T
        f.root.data.append(a)
    f.close()
    t.close()
    print("Completed", args.study_name)
    print("You can delete all other h5 files if genotype.h5 is correct")
    args.outfolder = args.genotype


def transpose_genotype_scheduler(args):
    local_run = False

    step4_name = args.genotype + '/' + args.study_name + '_step4_genotype_selected_variants.h5'
    step3_name = args.genotype + '/' + args.study_name + '_step3_genotype_no_missing.h5'
    step2_name = args.genotype + '/' + args.study_name + '_step2_merged_genotype.h5'
    
    if (os.path.exists(step4_name)):
        t = tables.open_file(step4_name, mode='r')
    elif (os.path.exists(step3_name)):
        print('using', step3_name)
        t = tables.open_file(step3_name, mode='r')
    elif (os.path.exists(step2_name)):
        print('using', step2_name)
        t = tables.open_file(step2_name, mode='r')
    else: 
        print('no valid genotype found')

    data = t.root.data
    num_pat = data.shape[1]
    t.close()
    print("n_jobs", args.n_jobs)
    jobchunk = int(np.ceil(num_pat / args.n_jobs))

    print("____________________________________________________________________")
    print('Submitting' + str(args.n_jobs))
    print('Please make sure this file has the correct settings for your cluster')
    print("____________________________________________________________________")
    with open('./gennet/utils/submit_SLURM_multi_transpose.sh', 'r') as f:
        print(f.read())
    print("____________________________________________________________________")

    if query_yes_no(question="Does the file contain the right settings?"):
        for job_n in (range(int(np.ceil(num_pat / jobchunk)))):
            begins = job_n * jobchunk
            tills = min(((job_n + 1) * jobchunk), num_pat)
            if local_run:
                # transpose_genotype_job(begins, tills, job_n, args.study_name, args.outfolder, args.tcm)
                str_sbatch = 'nohup python ./gennet/utils/Convert.py -job_begins ' + str(begins) + ' -job_tills ' + str(
                    tills) + ' -job_n ' + str(job_n) + ' -study_name ' + str(args.study_name) + ' -outfolder ' + str(
                    args.outfolder) + ' -tcm ' + str(args.tcm) + " & "
                print(str_sbatch)
                os.system(str_sbatch)
            else:
                str_sbatch = 'sbatch ./gennet/utils/submit_SLURM_multi_transpose.sh ' + str(begins) + ' ' + str(
                    tills) + ' ' + str(job_n) + ' ' + str(args.study_name) + ' ' + str(
                    args.outfolder) + ' ' + str(args.tcm)
                print(str_sbatch)
                os.system(str_sbatch)
        print("all jobs submitted please run GenNet convert -step merge_transpose next")
    else:
        print("Please change the script")
        transpose_genotype_scheduler(args)


def transpose_genotype_job(job_begins, job_tills, job_n, study_name, outfolder, tcm, comp_level=9):
    print("job_n:", job_n, 'job_begins:', job_begins, 'job_tills:', job_tills)
    hdf5_name = '/' + study_name + '_step4_genotype_selected_variants.h5'
    if (os.path.exists(outfolder + hdf5_name)):
        t = tables.open_file(outfolder + hdf5_name, mode='r')
    else:
        print('using', outfolder + study_name + '_step3_genotype_no_missing.h5')
        t = tables.open_file(outfolder + study_name + '_step3_genotype_no_missing.h5', mode='r')

    data = t.root.data
    num_pat = data.shape[1]
    num_feat = data.shape[0]
    chunk = tcm // num_feat
    chunk = int(np.clip(chunk, 1, num_pat))
    print("chuncksize =", chunk)

    f = tables.open_file(outfolder + study_name + '_step5_genotype_transposed_' + str(job_n) + '.h5', mode='w')
    f.create_earray(f.root, 'data', tables.IntCol(), (0, num_feat), expectedrows=num_pat,
                    filters=tables.Filters(complib='zlib', complevel=comp_level))
    f.close()
    n_in_job = job_tills - job_begins
    f = tables.open_file(outfolder + study_name + '_step5_genotype_transposed_' + str(job_n) + '.h5', mode='a')

    for subjects in tqdm.tqdm(range(int(np.ceil(n_in_job / chunk) + 1))):
        begins = job_begins + subjects * chunk
        tills = min((job_begins + (subjects + 1) * chunk), job_tills)
        a = np.array(data[:, begins:tills], dtype=np.int8)
        a = a.T
        f.root.data.append(a)
    f.close()
    t.close()
    print("Completed", job_n)


def merge_transpose(args):
    hdf5_name = '/' + args.study_name + '_step4_genotype_selected_variants.h5'
    if (os.path.exists(args.genotype + hdf5_name)):
        t = tables.open_file(args.genotype + hdf5_name, mode='r')
    else:
        print('using', args.genotype + args.study_name + '_step3_genotype_no_missing.h5')
        t = tables.open_file(args.genotype + args.study_name + '_step3_genotype_no_missing.h5', mode='r')

    num_pat = t.root.data.shape[1]
    num_feat = t.root.data.shape[0]
    chunk = args.tcm // num_feat
    chunk = int(np.clip(chunk, 1, num_pat))
    t.close()

    number_of_files = len(glob.glob(args.genotype + args.study_name + "_step5_genotype_transposed_*.h5"))

    if number_of_files == args.n_jobs:
        print('number of files ', number_of_files)
    else:
        print("WARNING!", 'number_of_files', number_of_files, 'args.n_jobs', args.n_jobs)
        print("Continueing to merge with n_jobs, merging:", args.n_jobs, "files")

    f = tables.open_file(args.outfolder + '/genotype.h5', mode='w')
    f.create_earray(f.root, 'data', tables.IntCol(), (0, num_feat), expectedrows=num_pat,
                    filters=tables.Filters(complib='zlib', complevel=args.comp_level))
    f.close()

    f = tables.open_file(args.outfolder + '/genotype.h5', mode='a')

    gen_tmp = tables.open_file(args.genotype + args.study_name + '_step5_genotype_transposed_' + str(0) + '.h5', mode='r')
    filesize = gen_tmp.root.data.shape[0]
    gen_tmp.close()
    print("\n merge all files...")
    if chunk > filesize:
        print("chunking is not necessary")
        for job_n in tqdm.tqdm(range(args.n_jobs)):
            gen_tmp = tables.open_file(args.genotype + args.study_name + '_step5_genotype_transposed_' + str(job_n) + '.h5', mode='r')
            f.root.data.append(np.array(np.round(gen_tmp.root.data[:, :]), dtype=int))
            gen_tmp.close()
        f.close()
    else:
        print("per chunk", chunk, "subjects")
        for job_n in tqdm.tqdm(range(args.n_jobs)):
            gen_tmp = tables.open_file(args.genotype + args.study_name + '_step5_genotype_transposed_' + str(job_n) + '.h5', mode='r')
            for chunckblock in range(int(np.ceil(gen_tmp.root.data.shape[0] / chunk))):
                begins = chunckblock * chunk
                tills = min(((chunckblock + 1) * chunk), gen_tmp.root.data.shape[0])
                f.root.data.append(np.array(np.round(gen_tmp.root.data[begins:tills, :]), dtype=int))
            gen_tmp.close()
        f.close()
    print("completed")
    f = tables.open_file(args.outfolder + '/genotype.h5', mode='r')
    finalshape = f.root.data.shape
    f.close()
    args.outfolder = args.genotype
    print("Shape merged file", finalshape)
    print("You can delete all other h5 files if genotype.h5 is correct")


def exclude_variants_probes(args):
    if args.variants is None:
        return

    used_indices = pd.read_csv(args.variants, header=None)
    used_indices = used_indices.index.values[used_indices.values.flatten()]
    probes = pd.read_hdf(args.genotype + '/probes/' + args.study_name + '.h5', mode="r")
    print("Probes shape", probes.shape)
    print("Selecting variants..")
    probes = probes.iloc[used_indices]
    print("Probes shape", probes.shape)
    probes.to_hdf(args.outfolder + '/probes/' + args.study_name + '_step4_genotype_selected_variants.h5', key='probes', format='table',
                  data_columns=True, append=True,
                  complib='zlib', complevel=args.comp_level, min_itemsize=45)
    args.outfolder = args.genotype


def getsum_(path_file):
    h5file = tables.open_file(path_file, mode="r")
    rowsum = 0
    for i in range(h5file.root.data.shape[0]):
        rowsum += np.sum(h5file.root.data[i, :])
    return rowsum


def get_checksum(path_file1, path_file2):
    sumfile1 = getsum_(path_file1)
    sumfile2 = getsum_(path_file2)

    print(sumfile1)
    print(sumfile2)
    assert sumfile1 == sumfile2


def select_first_arg_out(args):
    if type(args.out) is list:
        args.outfolder = args.out[0] + '/'
    else:
        args.outfolder = args.out + '/'


def select_first_arg_study(args):
    if type(args.study_name) is list:
        args.study_name = args.study_name[0]
    else:
        args.study_name = args.study_name

def check_genotype_folder(args):
    if args.genotype is str:
        pass
    else:
        args.genotype = args.outfolder
        

def convert(args):
    select_first_arg_out(args)

    if args.step == "all":
        # 1. hase
        hase_convert(args)
        # 2. merge
        select_first_arg_study(args)
        select_first_arg_out(args)
        check_genotype_folder(args)
        merge_hdf5_hase(args)
        # 3. impute
        impute_hase_hdf5(args)
        # 4. exclude variants
        exclude_variants_probes(args)
        # 5. transpose
        transpose_genotype(args)


    elif args.step == "hase_convert":
        hase_convert(args)
    elif args.step == "merge":
        check_genotype_folder(args)
        select_first_arg_study(args)
        merge_hdf5_hase(args)
    elif args.step == "impute_missing":
        check_genotype_folder(args)
        select_first_arg_study(args)
        impute_hase_hdf5(args)
    elif args.step == "exclude":
        check_genotype_folder(args)
        select_first_arg_study(args)
        exclude_variants_probes(args)
    elif ((args.step == "transpose") & (args.n_jobs == 1)):
        check_genotype_folder(args)
        select_first_arg_study(args)
        transpose_genotype(args)
    elif ((args.step == "transpose") & (args.n_jobs > 1)):
        check_genotype_folder(args)
        select_first_arg_study(args)
        transpose_genotype_scheduler(args)
    elif ((args.step == "merge_transpose")):
        check_genotype_folder(args)
        select_first_arg_study(args)
        merge_transpose(args)
    elif ((args.step == "checksum")):
        check_genotype_folder(args)
        print("Checksum does not count up if you exclude variants")
        select_first_arg_study(args)
        get_checksum(path_file1=args.outfolder + '/genotype.h5',
                     path_file2=args.outfolder + args.study_name + '_step2_merged_genotype.h5')

    else:
        print('invalid parameters')
        exit()


if __name__ == '__main__':
    CLI = argparse.ArgumentParser()
    CLI.add_argument(
        "-job_begins",
        type=int,
    )
    CLI.add_argument(
        "-job_tills",
        type=int,
    )
    CLI.add_argument(
        "-job_n",
        type=int,
    )
    CLI.add_argument(
        "-study_name",
        type=str,
        default=32,
    )
    CLI.add_argument(
        "-outfolder",
        type=str
    )
    CLI.add_argument(
        "-tcm",
        type=int,
    )
    arg = CLI.parse_args()
    transpose_genotype_job(job_begins=arg.job_begins,
                           job_tills=arg.job_tills,
                           job_n=arg.job_n,
                           study_name=arg.study_name,
                           outfolder=arg.outfolder,
                           tcm=arg.tcm)



import numpy as np

from gennet.utils.hase.hdgwas.tools import Timer


# @timing
def A_covariates(covariates, intercept=True):
    '''
    :param covariates: (n_subjects, n_covariates) - only constant covariates should be included (age, sex, ICV etc)
    :param intercept: default True, add intercept to model
    :return: matrix (n_cavariates, n_covariates), constant part for the rest of the study
    '''

    S, N = covariates.shape
    if intercept:
        I = np.ones(S).reshape(S, 1)
        covariates = np.hstack((I, covariates))
    a_cov = np.dot(covariates.T, covariates)
    return a_cov


# @timing
def B4(phenotype, genotype):
    b4 = np.tensordot(genotype, phenotype, axes=([1], [0]))
    return b4


def interaction(genotype, factor):
    g = genotype * factor.T
    return g


# @timing
def A_tests(covariates, genotype, intercept=True):  # TODO (low) extend for any number of tests in model
    '''
    :param covariates: (n_subjects, n_covariates) - only constant covariates should be included (age, sex, ICV etc)
    :param genotype: (n_tests, n_subjects) - test could be any kind of quantitative covariance
    :return: (1,n_covariates + intercept)
    '''

    if intercept:
        fst = np.sum(genotype, axis=1).reshape(-1, 1)
        sec = np.dot(genotype, covariates)
        tr = np.sum(np.power(genotype, 2), axis=1).reshape(-1, 1)
        return np.hstack((fst, sec, tr))

    else:
        sec = np.dot(genotype, covariates)
        tr = np.sum(np.power(genotype, 2), axis=1).reshape(-1, 1)
        return np.hstack((sec, tr))


# @timing
def B_covariates(covariates, phenotype, intercept=True):
    S, N = covariates.shape

    b_cov = np.dot(covariates.T, phenotype)
    if intercept:
        b1 = np.sum(phenotype, axis=0).reshape(1, phenotype.shape[1])
        B13 = np.append(b1, b_cov, axis=0)
        return B13
    else:
        return b_cov


# @timing
def A_inverse(a_covariates, a_test):  # TODO (low) extend for any number of tests in model

    A_inv = []
    n, m = a_covariates.shape
    k = n + 1
    for i in range(a_test.shape[0]):  # TODO (low) not in for loop
        inv = np.zeros(k * k).reshape(k, k)
        inv[0:k - 1, 0:k - 1] = a_covariates
        inv[k - 1, :] = a_test[i, :]
        inv[0:k, k - 1] = a_test[i, 0:k]
        try:
            A_inv.append(np.linalg.inv(inv))
        except:
            A_inv.append(np.zeros(k * k).reshape(k, k))  # TODO (high) test; check influence on results; warning;

    return np.array(A_inv)


# @timing
def C_matrix(phenotype):
    C = np.einsum('ij,ji->i', phenotype.T, phenotype)
    return C


# @timing
# @save_parameters
def HASE(b4, A_inverse, b_cov, C, N_con, DF):
    with Timer() as t:
        B13 = b_cov
        B4 = b4

        A1_B_constant = np.tensordot(A_inverse[:, :, 0:(N_con)], B13, axes=([2], [0]))

        A1_B_nonconstant = np.einsum('ijk,il->ijl', A_inverse[:, :, N_con:N_con + 1], B4)

        A1_B_full = A1_B_constant + A1_B_nonconstant

        BT_A1B_const = np.einsum('ij,lji->li', B13.T, A1_B_full[:, 0:(N_con), :])

        BT_A1B_nonconst = np.einsum('ijk,ijk->ijk', B4[:, None, :], A1_B_full[:, (N_con):N_con + 1, :])

        BT_A1B_full = BT_A1B_const[:, None, :] + BT_A1B_nonconst

        C_BTA1B = BT_A1B_full - C.reshape(1, -1)

        C_BTA1B = np.abs(C_BTA1B)

        a44_C_BTA1B = C_BTA1B * A_inverse[:, (N_con):N_con + 1, (N_con):N_con + 1]

        a44_C_BTA1B = np.sqrt((a44_C_BTA1B))

        t_stat = np.sqrt(DF) * np.divide(A1_B_full[:, (N_con):N_con + 1, :], a44_C_BTA1B)

        SE = a44_C_BTA1B / np.sqrt(DF)

    print("time to compute GWAS for {} phenotypes and {} SNPs .... {} sec".format(b4.shape[1],
                                                                                  A_inverse.shape[0],
                                                                                  t.secs))
    return t_stat, SE



import os
import sys
import os
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import plotly.express as px
import seaborn as sns
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
from gennet.utils.Utility_functions import get_paths

def plot_loss_function(resultpath):
        log_file = pd.read_csv(resultpath + "/train_log.csv")
        plt.plot(log_file['loss'])
        plt.plot(log_file['val_loss'])
        plt.title('Loss curve')
        plt.ylabel('loss')
        plt.xlabel('epoch')
        plt.legend(['train', 'val'], loc='upper right')
        plt.savefig(resultpath + "train_val_loss.png")
        plt.show()

def sunburst_plot(resultpath, importance_csv, num_layers=3, plot_threshold=0.01, add_end_node=True):
    csv_file = importance_csv.copy()

    number_of_weights = csv_file.filter(like="weights").shape[1]

    if add_end_node:
        csv_file["node_layer_" + str(number_of_weights)] = 0
        csv_file["layer" + str(number_of_weights) + '_name'] = "Prediction"

    plot_layer_names = []
    for i in range(number_of_weights, number_of_weights - num_layers, -1):
        plot_layer_names.append("layer" + str(i) + '_name')

    first_number_layer_to_plot = i

    csv_file["percentage_0"] = csv_file.filter(like="weights").prod(axis=1)
    csv_file["percentage_0"] = abs(csv_file["percentage_0"]) / abs(csv_file["percentage_0"]).sum() * 100
    print(csv_file["percentage_" + str(0)].sum())

    for i in range(1, first_number_layer_to_plot + 1):
        csv_file["percentage_" + str(i)] = csv_file.groupby("layer" + str(i) + '_name')[
            "percentage_" + str(i - 1)].transform('sum')

    print(csv_file["percentage_" + str(first_number_layer_to_plot)].sum())
    csv_file = csv_file.drop_duplicates("layer" + str(first_number_layer_to_plot) + '_name')
    print(csv_file["percentage_" + str(first_number_layer_to_plot)].sum())

    print(csv_file.shape)

    csv_file = csv_file[csv_file["percentage_" + str(first_number_layer_to_plot)] > plot_threshold]

    print(csv_file.shape)
    print(csv_file["percentage_" + str(first_number_layer_to_plot)].sum())
    print("start plotting")
    fig = px.sunburst(csv_file,
                      path=plot_layer_names,
                      values="percentage_" + str(first_number_layer_to_plot),
                      width=1000, height=1000,
                      template="presentation",
                      color_discrete_sequence=px.colors.qualitative.G10
                      )

    fig.write_image(resultpath + "sunburst.png")
    fig.write_html(resultpath + "sunburst.html")
    print(resultpath)
    print("done")


def plot_layer_weight(resultpath, importance_csv, layer=0, num_annotated=10):
    csv_file = importance_csv.copy()
    if int(layer) < csv_file.filter(like="weights").shape[1] - 1:
        pass
    elif int(layer) == csv_file.filter(like="weights").shape[1] - 1:
        csv_file["node_layer_" + str(csv_file.filter(like="weights").shape[1])] = 0
        csv_file["layer" + str(csv_file.filter(like="weights").shape[1]) + '_name'] = "end_node"
    else:
        print("error cant plot that many layers, there are not that many layers")
        sys.exit()

    if 'chr' in csv_file.columns:
        columns = ["node_layer_" + str(layer), "node_layer_" + str(layer + 1), "weights_" + str(layer),
                   'layer' + str(layer) + '_name', 'layer' + str(layer + 1) + '_name', 'chr']
        csv_file = csv_file[columns]
        csv_file = csv_file.drop_duplicates()

        csv_file = csv_file.sort_values(by=['chr', 'node_layer_' + str(layer)], ascending=True)
    else:
        columns = ["node_layer_" + str(layer), "node_layer_" + str(layer + 1), "weights_" + str(layer),
                   'layer' + str(layer) + '_name', 'layer' + str(layer + 1) + '_name']
        csv_file = csv_file[columns]
        csv_file = csv_file.drop_duplicates()

        csv_file = csv_file.sort_values(by=['node_layer_' + str(layer)], ascending=True)

    csv_file["pos"] = np.arange(len(csv_file))
    weights = abs(csv_file["weights_" + str(layer)].values)
    weights = weights / max(weights)
    csv_file["plot_weights"] = weights

    plt.figure(figsize=(20, 10))

    gene_middle = []
    if "chr" in csv_file.columns:
        color_end = np.sort(csv_file.groupby("chr")["pos"].max().values)
        print('coloring per chromosome')
        color_end = np.insert(color_end, 0, 0)
        for i in range(len(color_end) - 1):
            gene_middle.append((color_end[i] + color_end[i + 1]) / 2)
    else:
        color_end = np.sort(csv_file.groupby("node_layer_" + str(layer + 1))["pos"].max().values)
        color_end = np.insert(color_end, 0, 0)
        print("no chr information continuing by coloring per group in node_layer_1")

    colormap = ['#7dcfe2', '#4b78b5', 'darkgrey', 'dimgray'] * len(color_end)

    if len(weights) < 500:

        sns.set(style="whitegrid")
        sns.set_color_codes("pastel")

        f, ax = plt.subplots(figsize=(6, 10))

        sns.barplot(x="plot_weights", y='layer' + str(layer) + '_name', data=csv_file,
                    label="Total")

        ax.set(ylabel="Layer node",
               xlabel="Normalized Weights")
        new_labels = [label for label in ax.get_yticklabels()]
        fontsize = int(np.round((-8 / 90) * len(new_labels) + 13.888))
        ax.set_yticklabels(new_labels, fontsize=fontsize)
        sns.despine(left=True, bottom=True)
        plt.savefig(resultpath + "manhattan_weights_" + str(layer) + ".png", bbox_inches='tight')

    else:
        for i in range(len(color_end) - 1):
            plt.scatter(csv_file['pos'].iloc[color_end[i]:color_end[i + 1]],
                        csv_file["plot_weights"].iloc[color_end[i]:color_end[i + 1]], c=colormap[i])

        plt.ylim(bottom=0, top=1.2)
        plt.xlim(0, len(weights) + int(len(weights) / 100))
        plt.title("Network Weights layer " + str(layer), size=36)
        if len(gene_middle) > 1:
            plt.xticks(gene_middle, np.arange(len(gene_middle)) + 1, size=16)
            plt.xlabel("Chromosome", size=18)
        else:
            plt.xlabel("Chromosome position", size=18)
        plt.ylabel("Weights", size=18)

        gene5_overview = csv_file.sort_values("plot_weights", ascending=False).head(num_annotated)

        if len(gene5_overview) < num_annotated:
            num_annotated = len(gene5_overview)
        for i in range(num_annotated):
            print(gene5_overview['layer' + str(layer) + '_name'].iloc[i],
                  gene5_overview["pos"].iloc[i], gene5_overview["plot_weights"].iloc[i])
            plt.annotate(gene5_overview['layer' + str(layer) + '_name'].iloc[i],
                         (gene5_overview["pos"].iloc[i], gene5_overview["plot_weights"].iloc[i]),
                         xytext=(gene5_overview["pos"].iloc[i],
                                 gene5_overview["plot_weights"].iloc[i]), size=16)

        plt.gca().spines['right'].set_color('none')
        plt.gca().spines['top'].set_color('none')

        plt.savefig(resultpath + "Weights_layer_" + str(layer) + ".png", bbox_inches='tight', pad_inches=0)


def manhattan_relative_importance(resultpath, importance_csv, num_annotated=10):
    csv_file = importance_csv.copy()
    plt.figure(figsize=(20, 10))

    gene_middle = []


    if "chr" in csv_file.columns:
        csv_file = csv_file.sort_values(by=['chr', 'node_layer_0'], ascending=True)
        csv_file["pos"] = np.arange(len(csv_file))
        color_end = np.sort(csv_file.groupby("chr")["pos"].max().values+1)
        print('coloring per chromosome')
        color_end = np.insert(color_end, 0, 0)
        for i in range(len(color_end) - 1):
            gene_middle.append((color_end[i] + color_end[i + 1]) // 2)
    else:
        csv_file = csv_file.sort_values(by=['node_layer_0'], ascending=True)
        csv_file["pos"] = np.arange(len(csv_file))
        color_end = np.sort(csv_file.groupby("node_layer_1")["pos"].max().values)
        color_end = np.insert(color_end, 0, 0)
        print("no chr information continuing by coloring per group in node_layer_1")

    weights = abs(csv_file["raw_importance"])
    weights = weights / max(weights)
    csv_file["plot_weights"] = weights

    print(len(color_end), "color groups")
    colormap = ['#7dcfe2', '#4b78b5', 'darkgrey', 'dimgray'] * len(color_end)

    for i in range(len(color_end) - 1):
        plt.scatter(csv_file['pos'].iloc[color_end[i]:color_end[i + 1]],
                    csv_file["plot_weights"].iloc[color_end[i]:color_end[i + 1]], c=colormap[i])

    plt.ylim(bottom=0, top=1.2)
    plt.xlim(0, len(weights) + int(len(weights) / 100))
    plt.title("Relative importance of all SNPs", size=36)
    if len(gene_middle) > 1:
        plt.xticks(gene_middle, np.unique(csv_file["chr"]), size=16)
        plt.xlabel("Chromosome", size=18)
    else:
        plt.xlabel("Chromosome position", size=18)
    plt.ylabel("Relative importance", size=18)

    offset = len(csv_file) / 200
    offset = np.clip(offset, 0.1, 100)

    gene5_overview = csv_file.sort_values("plot_weights", ascending=False).head(num_annotated)

    if len(gene5_overview) < num_annotated:
        num_annotated = len(gene5_overview)
    for i in range(num_annotated):
        plt.annotate(gene5_overview['layer0_name'].iloc[i],
                     (gene5_overview["pos"].iloc[i], gene5_overview["plot_weights"].iloc[i]),
                     xytext=(gene5_overview["pos"].iloc[i] + offset,
                             gene5_overview["plot_weights"].iloc[i]), size=16)

    plt.gca().spines['right'].set_color('none')
    plt.gca().spines['top'].set_color('none')

    plt.savefig(resultpath + "Manhattan_relative_importance_SNPs.png", bbox_inches='tight', pad_inches=0)


def plot(args):
    folder, resultpath = get_paths(args)
    importance_csv = pd.read_csv(resultpath + "/connection_weights.csv", index_col=0)
    print(resultpath)
    layer = args.layer_n
    if args.type == "layer_weight":
        plot_layer_weight(resultpath, importance_csv, layer=layer, num_annotated=10)
    elif args.type == "sunburst":
        sunburst_plot(resultpath=resultpath, importance_csv=importance_csv)
    elif args.type == "manhattan_relative_importance":
        manhattan_relative_importance(resultpath=resultpath, importance_csv=importance_csv)
    else:
        print("invalid type:", args.type)
        exit()



import numpy as np
import tqdm
import shap


def DFIM_test_index(explainer, xtest, index_snps):
    
    # Lists to hold the maximum differences
    perturbed_values = []
    max_not_perturbed = []
    loc_not_perturbed = []
    
    onehot = False if len(xtest.shape) == 2 else True

    for i in tqdm.tqdm(index_snps):  # Assuming second dimension of xtest is features
        variant_shap_values = []

        if onehot == True:
            for SNP_perm in range(3):
                xtest_copy = xtest.copy()
                xtest_copy[:,i, :] = np.zeros(3)
                xtest_copy[:,i, SNP_perm] = 1
        
                shap_values = np.max(explainer.shap_values(xtest_copy)[0], axis=2)
                variant_shap_values.append(shap_values)
        else:
            for SNP_perm in range(3):
                xtest_copy = xtest.copy()
                xtest_copy[:, i] = SNP_perm
                shap_values = explainer.shap_values(xtest_copy)[0]
                variant_shap_values.append(shap_values)

        # Calculate differences between the perturbations for the i-th feature
        differences = []
        for idx1 in range(len(variant_shap_values)):
            for idx2 in range(idx1 + 1, len(variant_shap_values)):
                differences.append(np.abs(variant_shap_values[idx1] - variant_shap_values[idx2]))

        # Stack the differences to get a 3D array of shape (number_of_comparisons, number_of_samples, number_of_features)
        stacked_differences = np.stack(differences)
        stacked_differences = np.mean(stacked_differences, axis=1)
        stacked_differences = np.max(stacked_differences, axis=0)

        # Find the max difference for the perturbed feature (i-th feature)
        perturbed_values.append(stacked_differences[i])

        # Find the max difference for all other features (not i-th feature)
        # We mask the i-th feature to exclude it from the calculation
        stacked_differences[i] = 0
        max_not_perturbed.append(np.max(stacked_differences))
        loc_not_perturbed.append(np.argmax(stacked_differences))

    
    return perturbed_values, max_not_perturbed, loc_not_perturbed





import glob
import os
import sys

sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

import matplotlib
import numpy as np
import pandas as pd
import tables
import tensorflow.keras as K

matplotlib.use('agg')


def check_data(datapath, genotype_path, mode):
    # TODO write checks for multiple genotype files.
    # global groundtruth # why is this a global? # removed did it break something?
    groundtruth = None
    genotype_matrix = False
    network_structure = False
    patient_info = False
    multiple_genotype_matrices = False
    number_of_covariats = False
    classification_problem = "undetermined"

    if os.path.exists(genotype_path + 'genotype.h5'):
        genotype_matrix = True
    elif len(glob.glob(genotype_path + '*.h5')) > 0:
        multiple_genotype_matrices = True
    else:
        print("genotype missing in", genotype_path)

    if os.path.exists(datapath + 'topology.csv'):
        network_structure = True
    elif len(glob.glob(datapath + '*.npz')) > 0:
        network_structure = True
    else:
        print("topology.csv and *.npz are missing")
    if os.path.exists(datapath + 'subjects.csv'):
        patient_info = True
        groundtruth = pd.read_csv(datapath + "/subjects.csv")
        
        number_of_covariats = groundtruth.filter(like="cov_").shape[1]
        print('number of covariates:', number_of_covariats)
        print('Covariate columns found:', list(groundtruth.filter(like="cov_").columns.values))
        
        if {'patient_id', 'labels', 'genotype_row', 'set'}.issubset(groundtruth.columns):
            classification_problem = ((groundtruth["labels"].values == 0) | (groundtruth["labels"].values == 1)).all()
        else:
            print("column names missing need 'patient_id', 'labels', 'genotype_row', 'set', got:",
                  groundtruth.columns.values)
            exit()
    else:
        print("subjects.csv is missing")

    print("mode is", mode)
    
    if (mode == "classification") and classification_problem:
        pass
    elif (mode == "regression") and not classification_problem:
        pass
    else:
        print("The labels and the given mode do not correspond. \n"
              "Classification problems should have binary labels [1, 0]. \n"
              "Regression problems should not be binary. \n"
              "The labels do have the following values", groundtruth["labels"].unique())
        exit()

    if multiple_genotype_matrices & network_structure & patient_info:
        TrainDataGenerator.multi_h5 = True
        return
    if genotype_matrix & network_structure & patient_info:
        return
    else:
        print("Did you forget the last (/) slash?")
        exit()


def get_inputsize(genotype_path):
    single_genotype_path = glob.glob(genotype_path + '*.h5')[0]
    h5file = tables.open_file(single_genotype_path, "r")
    inputsize = h5file.root.data.shape[1]
    h5file.close()
    return inputsize


def get_labels(datapath, set_number):
    groundtruth = pd.read_csv(datapath + "/subjects.csv")
    groundtruth = groundtruth[groundtruth["set"] == set_number]
    ybatch = np.reshape(np.array(groundtruth["labels"].values), (-1, 1))
    return ybatch




class TrainDataGenerator(K.utils.Sequence):

    def __init__(self, datapath, genotype_path, batch_size, trainsize, inputsize, epoch_size, shuffle=True, one_hot=False):
        self.datapath = datapath
        self.batch_size = batch_size
        self.genotype_path = genotype_path
        self.shuffledindexes = np.arange(trainsize)
        self.trainsize = trainsize
        self.multi_h5 = len(glob.glob(self.genotype_path + '*.h5')) > 1
        self.h5filenames = "_UKBB_MRI_QC_T_M"
        self.training_subjects = pd.read_csv(self.datapath + "/subjects.csv")
        self.training_subjects = self.training_subjects[self.training_subjects["set"] == 1]
        self.inputsize = inputsize
        self.epoch_size = epoch_size
        self.left_in_greater_epoch = trainsize
        self.count_after_shuffle = 0
        self.one_hot = one_hot

        if shuffle:
            np.random.shuffle(self.shuffledindexes)

    def __len__(self):
        return int(np.ceil(self.epoch_size / float(self.batch_size)))
        

    def __getitem__(self, idx):
        if self.multi_h5:
            xbatch, ybatch = self.multi_genotype_matrix(idx)
        else:
            xbatch, ybatch = self.single_genotype_matrix(idx)

        return xbatch, ybatch


    def if_one_hot(self, xbatch):       
        xbatch_dim = len(xbatch.shape) 
        if self.one_hot:
            if xbatch_dim == 3:
                pass
            elif xbatch_dim == 2:
                xbatch = K.utils.to_categorical(np.array(xbatch, dtype=np.int8))
            else:
                print("unexpected shape!")   
        return xbatch
    
    
    def single_genotype_matrix(self, idx):
        idx2 = idx + self.count_after_shuffle      
        genotype_hdf = tables.open_file(self.genotype_path + "/genotype.h5", "r")
        batchindexes = self.shuffledindexes[idx2 * self.batch_size:((idx2 + 1) * self.batch_size)]
        ybatch = self.training_subjects["labels"].iloc[batchindexes]
        xcov = self.training_subjects.filter(like="cov_").iloc[batchindexes]
        xcov = xcov.values
        xbatchid = np.array(self.training_subjects["genotype_row"].iloc[batchindexes], dtype=int)
        xbatch = genotype_hdf.root.data[xbatchid, :] 
        xbatch = self.if_one_hot(xbatch)
        ybatch = np.reshape(np.array(ybatch), (-1, 1))
        genotype_hdf.close()
        return [xbatch, xcov], ybatch

    def multi_genotype_matrix(self, idx):
        idx2 = idx + self.count_after_shuffle 
        batchindexes = self.shuffledindexes[idx2 * self.batch_size:((idx2 + 1) * self.batch_size)]
        ybatch = self.training_subjects["labels"].iloc[batchindexes]
        
        xcov = self.training_subjects.filter(like="cov_").iloc[batchindexes]
        xcov = xcov.values
        
        subjects_current_batch = self.training_subjects.iloc[batchindexes]
        subjects_current_batch["batch_index"] = np.arange(len(subjects_current_batch))
        xbatch = np.zeros((len(ybatch), self.inputsize))
        for i in subjects_current_batch["chunk_id"].unique():
            genotype_hdf = tables.open_file(self.genotype_path + "/" + str(i) + self.h5filenames + ".h5", "r")
            subjects_current_chunk = subjects_current_batch[subjects_current_batch["chunk_id"] == i]
            xbatchid = np.array(subjects_current_chunk["genotype_row"].values, dtype=int)
            if len(xbatchid) > 1:
                pass
            else:
                xbatchid = int(xbatchid)
            xbatch[subjects_current_chunk["batch_index"].values, :] = genotype_hdf.root.data[xbatchid, :]
            genotype_hdf.close()
            
        xbatch = self.to_one_hot(xbatch)
        ybatch = np.reshape(np.array(ybatch), (-1, 1))
        return [xbatch, xcov], ybatch

    def on_epoch_end(self):
        """Updates indexes after each epoch"""
        left_in_epoch = self.left_in_greater_epoch - self.epoch_size
        print(left_in_epoch, 'left_in_epoch')
        if  left_in_epoch < self.epoch_size: 
            print("Shuffeling epochs")
            np.random.shuffle(self.shuffledindexes)
            self.left_in_greater_epoch = self.trainsize
            self.count_after_shuffle = 0
        else:
            self.left_in_greater_epoch = self.left_in_greater_epoch - self.epoch_size
            self.count_after_shuffle = self.count_after_shuffle + int(np.ceil(self.epoch_size / float(self.batch_size)))
            
          


class EvalGenerator(K.utils.Sequence):

    def __init__(self, datapath, genotype_path, batch_size, setsize, inputsize, evalset="undefined", one_hot=False):
        self.datapath = datapath
        self.batch_size = batch_size
        self.yvalsize = setsize
        self.inputsize = inputsize
        self.genotype_path = genotype_path
        self.h5file = []
        self.h5filenames = "_UKBB_MRI_QC_T_M"
        self.multi_h5 = len(glob.glob(self.genotype_path + '*.h5')) > 1
        self.eval_subjects = pd.read_csv(self.datapath + "/subjects.csv")
        self.one_hot = one_hot
        
        if evalset == "validation":
            self.eval_subjects = self.eval_subjects[self.eval_subjects["set"] == 2]
        elif evalset == "test":
            self.eval_subjects = self.eval_subjects[self.eval_subjects["set"] == 3]
        else:
            print("please add which evalset should be used in the call, validation or test. Currently undefined")

    def __len__(self):
        val_len = int(np.ceil(self.yvalsize / float(self.batch_size)))
        return val_len

    def __getitem__(self, idx):
        if self.multi_h5:
            xbatch, ybatch = self.multi_genotype_matrix(idx)
        else:
            xbatch, ybatch = self.single_genotype_matrix(idx)

        return xbatch, ybatch

    def if_one_hot(self, xbatch):       
        xbatch_dim = len(xbatch.shape) 
        if self.one_hot:
            if xbatch_dim == 3:
                pass
            elif xbatch_dim == 2:
                xbatch = K.utils.to_categorical(np.array(xbatch, dtype=np.int8))
            else:
                print("unexpected shape!")   
        return xbatch

    def single_genotype_matrix(self, idx):
        genotype_hdf = tables.open_file(self.genotype_path + "/genotype.h5", "r")
        ybatch = self.eval_subjects["labels"].iloc[idx * self.batch_size:((idx + 1) * self.batch_size)]
        xcov = self.eval_subjects.filter(like="cov_").iloc[idx * self.batch_size:((idx + 1) * self.batch_size)]
        xcov = xcov.values
        xbatchid = np.array(self.eval_subjects["genotype_row"].iloc[idx * self.batch_size:((idx + 1) * self.batch_size)],
                            dtype=int)
        xbatch = genotype_hdf.root.data[xbatchid, :]  
        xbatch = self.if_one_hot(xbatch)
        ybatch = np.reshape(np.array(ybatch), (-1, 1))
        genotype_hdf.close()
        return [xbatch, xcov], ybatch

    def multi_genotype_matrix(self, idx):      
        subjects_current_batch = self.eval_subjects.iloc[idx * self.batch_size:((idx + 1) * self.batch_size)]
        subjects_current_batch["batch_index"] = np.arange(subjects_current_batch.shape[0])
              
        xbatch = np.zeros((len(subjects_current_batch["labels"]), self.inputsize))

        for i in subjects_current_batch["chunk_id"].unique():
            genotype_hdf = tables.open_file(self.genotype_path + "/" + str(i) + self.h5filenames + ".h5", "r")
            subjects_current_chunk = subjects_current_batch[subjects_current_batch["chunk_id"] == i]
            xbatchid = np.array(subjects_current_chunk["genotype_row"].values, dtype=int)
            xbatch[subjects_current_chunk["batch_index"].values, :] = genotype_hdf.root.data[xbatchid, :]
            genotype_hdf.close()
            
        xbatch = self.to_one_hot(xbatch)
        ybatch = np.reshape(np.array(subjects_current_batch["labels"]), (-1, 1))
        xcov = subjects_current_batch.filter(like="cov_").values                     
        return [xbatch, xcov], ybatch
    

    def get_data(self, sample_pat=0):

        genotype_hdf = tables.open_file(self.genotype_path + "/genotype.h5", "r")
        ybatch = self.eval_subjects["labels"]

        if sample_pat > 0:
            self.eval_subjects = self.eval_subjects.sample(n=sample_pat, random_state=1)
        
        xbatchid = np.array(self.eval_subjects["genotype_row"].values, dtype=int)
            
        xcov = self.eval_subjects.filter(like="cov_")
        xcov = xcov.values
        xbatch = genotype_hdf.root.data[xbatchid,...]  
        xbatch = self.if_one_hot(xbatch)
        ybatch = np.reshape(np.array(ybatch), (-1, 1))
        genotype_hdf.close()
        return [xbatch, xcov], ybatch





import argparse
import gc
import os

import numpy as np
import pandas as pd

parser = argparse.ArgumentParser(description='Make phenotype folder ready for HASE analysis')
parser.add_argument("-i", required=True, type=str, help="path to nparrays")
parser.add_argument("-id", type=str, help="csv data frame with two columns:"
                                          "1)id: ids identical to ids in genotype data"
                                          "2) exclude: 1 if exclude, 0 if include to analysis")

args = parser.parse_args()
print(args)

df = pd.read_csv(args.id, index_col=0)

print(df.head())
subject_id = []

for i in df.iterrows():
    if i[1].exclude == 1:
        subject_id.append('remove_' + str(i[1].id))
    else:
        subject_id.append(str(i[1].id))

dic = {}
dic['id'] = subject_id

files = os.listdir(args.i)

for f in files:
    print(f)
    if f.split('.')[-1] != 'npy':
        raise ValueError('In {} should be only nparrays, not {}'.format(args.i, f))

    d = np.load(os.path.join(args.i, f))
    ch = f.split('_')[0].split('reg')[1]
    index = f.split('_')[1].split('.npy')[0]
    n, m = d.shape
    if n != len(subject_id):
        raise ValueError(
            'Number of ids {} from data frame not equal to number of ids {} from file {} '.format(len(subject_id), n,
                                                                                                  f))
    dic[f] = [ch + "_" + index + "_" + str(i) for i in range(m)]
    d = None
    gc.collect()

np.save(os.path.join(args.i, 'info_dic.npy'), dic)



# Locally-Directed1D layer
#
# For the article see: https://www.biorxiv.org/content/10.1101/2020.06.19.159152v1
# For an explanation how to use this layer see https://github.com/ArnovanHilten/GenNet
# Locallyconnected1D is used as a basis to write the LocallyDirected layer

"""Locally-Directed1D layer.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
import tensorflow as tf

from tensorflow import concat
from tensorflow.python.keras import backend as Kbb
from tensorflow.keras import backend as Kb
from tensorflow.keras import activations
from tensorflow.keras import constraints
from tensorflow.keras import initializers
from tensorflow.keras import regularizers
from tensorflow.keras.layers import Layer
from tensorflow.keras.layers import InputSpec
from tensorflow.python.keras.utils import conv_utils
from tensorflow.python.keras.utils import tf_utils
from tensorflow.python.util.tf_export import keras_export

@keras_export('keras.layers.LocallyConnected1D')
class LocallyDirected1D(Layer):
    """Locally-Directed1D layer for 1D inputs.

    Dense layer with custom connections. The custom connections are defined by the mask input, a sparse (COO) connectivity matrix.

    # The matrix has the shape of (N_nodes_layer_1, N_nodes_layer_2).
    # It is a sparse matrix with zeros for no connections and ones if there is a connections. For example.


    #             output
    #           1 2 3 4 5
    # input 1 | 1 0 0 0 0 |
    # input 2 | 1 1 0 0 0 |
    # input 3 | 0 1 0 0 0 |
    # input 4 | 0 1 0 0 0 |
    # input 5 | 0 0 1 0 0 |
    # input 6 | 0 0 0 1 0 |
    # input 7 | 0 0 0 1 0 |


    # This connects the first two inputs (1,2) to the first neuron in the second layer.
    # Connects input 2,3 and 4 to output neuron 2.
    # Connects input 5 to output neuron 3
    # Connects input 6 and 7 o the 4th neuron in the subsequent layer
    # Connects nothing to the 5th neuron
    #
    # Writtem for Gennet framework: interpretable neural networks for phenotype prediction
    # (https://www.biorxiv.org/content/10.1101/2020.06.19.159152v1.full)


  Arguments:
      mask: sparse matrix with shape (input, output) connectivity matrix,
            True defines connection between (in_i, out_j), should be sparse (False,0) >> True
            should be scipy sparese matrix in COO Format!
      filters: Integer, the dimensionality of the output space
          (i.e. the number of output filters in the convolution).
      padding: Currently only supports `"valid"` (case-insensitive).
          `"same"` may be supported in the future.
      data_format: A string,
          one of `channels_last` (default) or `channels_first`.
          The ordering of the dimensions in the inputs.
          `channels_last` corresponds to inputs with shape
          `(batch, length, channels)` while `channels_first`
          corresponds to inputs with shape
          `(batch, channels, length)`.
          It defaults to the `image_data_format` value found in your
          Keras config file at `~/.keras/keras.json`.
          If you never set it, then it will be "channels_last".
      activation: Activation function to use.
          If you don't specify anything, no activation is applied
          (ie. "linear" activation: `a(x) = x`).
      use_bias: Boolean, whether the layer uses a bias vector.
      kernel_initializer: Initializer for the `kernel` weights matrix.
      bias_initializer: Initializer for the bias vector.
      kernel_regularizer: Regularizer function applied to
          the `kernel` weights matrix.
      bias_regularizer: Regularizer function applied to the bias vector.
      activity_regularizer: Regularizer function applied to
          the output of the layer (its "activation")..
      kernel_constraint: Constraint function applied to the kernel matrix.
      bias_constraint: Constraint function applied to the bias vector.

  Input shape:
      3D tensor with shape: `(batch_size, steps, input_dim)`

  Output shape:
      3D tensor with shape: `(batch_size, new_steps, filters)`
      `steps` value might have changed due to padding or strides.
  """

    def __init__(self,
                 mask,
                 filters,
                 padding='valid',
                 data_format=None,
                 activation=None,
                 use_bias=True,
                 kernel_initializer='glorot_uniform',
                 bias_initializer='zeros',
                 kernel_regularizer=None,
                 bias_regularizer=None,
                 activity_regularizer=None,
                 kernel_constraint=None,
                 bias_constraint=None,
                 **kwargs):
        super(LocallyDirected1D, self).__init__(**kwargs)
        self.filters = filters
        self.padding = conv_utils.normalize_padding(padding)
        self.data_format = conv_utils.normalize_data_format(data_format)
        self.activation = activations.get(activation)
        self.use_bias = use_bias
        self.kernel_initializer = initializers.get(kernel_initializer)
        self.bias_initializer = initializers.get(bias_initializer)
        self.kernel_regularizer = regularizers.get(kernel_regularizer)
        self.bias_regularizer = regularizers.get(bias_regularizer)
        self.activity_regularizer = regularizers.get(activity_regularizer)
        self.kernel_constraint = constraints.get(kernel_constraint)
        self.bias_constraint = constraints.get(bias_constraint)
        self.input_spec = InputSpec(ndim=3)
        self.mask = mask
        self.mask_indices = []
        self.input_filters = None

    @tf_utils.shape_type_conversion
    def build(self, input_shape):
        if self.data_format == 'channels_first':
            input_dim, input_length = input_shape[1], input_shape[2]
        else:
            input_dim, input_length = input_shape[2], input_shape[1]

        if input_dim is None:
            raise ValueError('Axis 2 of input should be fully-defined. '
                             'Found shape:', input_shape)
            
        self.input_filters = input_dim
        self.output_length = self.mask.shape[1] 

        if self.data_format == 'channels_first':
            self.kernel_shape = (input_dim, input_length,
                                 self.filters, self.output_length)
        else:
            self.kernel_shape = (input_length, input_dim,
                                 self.output_length, self.filters)
        
        self.kernel = self.add_weight(shape=(len(self.mask.data), input_dim*self.filters ),    
        #sum of all nonzero values in mask sum(sum(mask))
                            initializer=self.kernel_initializer,
                            name='kernel',
                            regularizer=self.kernel_regularizer,
                            constraint=self.kernel_constraint)
                        
        self.mask_indices = np.array([self.mask.row, self.mask.col]).T

        if self.use_bias:
            self.bias = self.add_weight(
                shape=(self.output_length, self.filters * self.input_filters),
                initializer=self.bias_initializer,
                name='bias',
                regularizer=self.bias_regularizer,
                constraint=self.bias_constraint)
        else:
            self.bias = None

        if self.data_format == 'channels_first':
            self.input_spec = InputSpec(ndim=3, axes={1: input_dim})
        else:
            self.input_spec = InputSpec(ndim=3, axes={-1: input_dim})
        self.built = True

    def call(self, inputs):
        
        output_filters = []
   
        for input_filter_n in range(self.input_filters): 
            for i in range(self.filters):
                weight_matrix = tf.scatter_nd(self.mask_indices, self.kernel[:,i], (self.mask.shape[0], self.mask.shape[1]))
                output_mat = tf.matmul(inputs[:,:,input_filter_n], weight_matrix) # 
                output_mat = Kb.reshape(output_mat, [-1, self.output_length, 1])  
                output_filters.append(output_mat)

        
        output = concat(output_filters, axis=-1)   
        
        if self.use_bias:
            output = Kb.bias_add(output, self.bias, data_format=self.data_format)

        output = self.activation(output)
        return output
   

    def get_config(self):
        config = {
            # 'mask':  # replace by two numpy arrays with indices
            # self.mask,
            'filters':
                self.filters,
            'padding':
                self.padding,
            'data_format':
                self.data_format,
            'activation':
                activations.serialize(self.activation),
            'use_bias':
                self.use_bias,
            'kernel_initializer':
                initializers.serialize(self.kernel_initializer),
            'bias_initializer':
                initializers.serialize(self.bias_initializer),
            'kernel_regularizer':
                regularizers.serialize(self.kernel_regularizer),
            'bias_regularizer':
                regularizers.serialize(self.bias_regularizer),
            'activity_regularizer':
                regularizers.serialize(self.activity_regularizer),
            'kernel_constraint':
                constraints.serialize(self.kernel_constraint),
            'bias_constraint':
                constraints.serialize(self.bias_constraint),
        }
        base_config = super(LocallyDirected1D, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))





import os


class Protocol:  # TODO (mid) write class

    def __init__(self, path):
        self.name = None
        self.model = None
        self.covariates = None
        self.MAF = None
        self.QC = None
        self.genotype_format = None
        self.phenotype_format = None
        self.type = None  # meta,single etc
        self.enable = False
        self.path = path
        if os.path.isfile(self.path):
            self.enable = True
            self.parse()

    def parse(self):
        print('Not implemented!')

    def regression_model(self):
        pass



import argparse

parser = argparse.ArgumentParser(description='Script to generate data')

parser.add_argument('-type', type=str, required=True, choices=['phenotype'],
                    help=' Choose what kind of data do you want to generate')

parser.add_argument("-s", type=int, help="number of subject")
parser.add_argument("-ph", type=int, help="number of phenotypes")
parser.add_argument("-o", "--out", type=str, required=True, help="path to save result folder")
parser.add_argument("-save_name", type=str, required=True, help="file name to save")

args = parser.parse_args()
print(args)


def random_phenotype(Ns, Np):
    pass



from gennet.utils.hase.hdgwas.data import Reader


class HASE:
    def __init__(self):
        self.cluster = 'n'
        self.effect_intercept = False
        self.hdf5 = True
        self.id = False
        self.interaction = None
        self.intercept = 'y'
        self.maf = 0.0
        self.mapper_chunk = None
        self.node = None
        self.np = True
        self.pd_full = False
        self.permute_ph = False
        self.protocol = None
        self.ref_name = '1000Gp1v3_ref'
        self.thr = None
        self.vcf = False

        self.ph_id_exc = None
        self.ph_id_inc = None
        self.snp_id_inc = None
        self.snp_id_exc = None

        self.mapper = None
        self.mode = None

        self.derivatives = []
        self.encoded = []
        self.genotype = []
        self.phenotype = []
        self.covariates = []
        self.study_name = []

        # os.environ['HASEDIR'] = ''
        # os.environ['HASEOUT'] = ''

        self.flags = {}

    def add_study(self, study):

        if study.__class__.__name__ == 'Study':

            if study.study_name not in self.study_name:
                self.study_name.append(study.study_name)
                if study.derivatives is not None:
                    self.derivatives.append(study.derivatives)
                    if study.encoded is not None:
                        self.encoded.append(study.encoded)
                    else:
                        raise ValueError('There is no info about study encode status!')

                elif study.covariates is not None:
                    self.covariates.append(study.covariates)
                else:
                    raise ValueError('Study should have covariates or derivatives!')
                if study.genotype is not None:
                    self.genotype.append(study.genotype)
                if study.phenotype is not None:
                    self.phenotype.append(study.phenotype)


            else:
                raise ValueError('You already added study with name {}'.format(study.study_name))

        else:
            raise ValueError('You can add only Study object!')


class Study:
    def __init__(self, name):
        self.name = name
        self.genotype = None
        self.phenotype = None
        self.derivatives = None
        self.covariates = None
        self.study_name = None
        self.encoded = None

    def add_genotype(self, genotype_path, hdf5=True):
        self.genotype = Reader('genotype')
        self.genotype.start(genotype_path, hdf5=hdf5, study_name=self.study_name, ID=False)

    def add_phenotype(self, phenotype_path):
        self.phenotype = Reader('phenotype')
        self.phenotype.start(phenotype_path)

    def add_derivatives(self, derivatives_path):
        self.derivatives = Reader('partial')
        self.derivatives.start(derivatives_path, study_name=self.study_name)
        self.derivatives.folder.load()

    def add_covariates(self, covariates_path):
        self.covariates = Reader('covariates')
        self.covariates.start(covariates_path)
